# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
import pandas as pd

from ax.analysis.healthcheck.regression_analysis import RegressionAnalysis
from ax.core.data import Data
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment_with_multi_objective


class TestRegressionAnalysis(TestCase):
    def test_regression_analysis(self) -> None:
        experiment = get_branin_experiment_with_multi_objective(
            with_batch=True, with_status_quo=True
        )

        df = pd.DataFrame(
            {
                "metric_name": ["branin_a"] * 6 + ["branin_b"] * 6,
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"] * 2,
                "trial_index": [0] * 12,
                "mean": list(np.arange(7, 1, -1)) + list(np.arange(6, 1, -1)) + [10.0],
                "sem": [1.0] * 12,
            }
        )

        experiment.attach_data(Data(df=df))
        ra = RegressionAnalysis(prob_threshold=0.90)
        card = ra.compute(experiment=experiment, generation_strategy=None)
        self.assertEqual(card.name, "RegressionAnalysis")
        self.assertEqual(card.title, "Ax Regression Analysis Warning")
        self.assertTrue(
            card.subtitle is not None
            and "0_4" in card.subtitle
            and "branin_b" in card.subtitle
            and "Trial 0" in card.subtitle
        )

        df = pd.DataFrame(
            {
                "metric_name": ["branin_a"] * 6 + ["branin_b"] * 6,
                "arm_name": ["status_quo", "0_0", "0_1", "0_2", "0_3", "0_4"] * 2,
                "trial_index": [0] * 12,
                "mean": list(np.arange(7, 1, -1)) * 2,
                "sem": [1.0] * 12,
            }
        )
        experiment.attach_data(Data(df=df))
        ra = RegressionAnalysis(prob_threshold=0.90)
        card = ra.compute(experiment=experiment, generation_strategy=None)
        self.assertEqual(card.name, "RegressionAnalysis")
        self.assertEqual(card.title, "Ax Regression Analysis Success")
        self.assertEqual(
            card.subtitle,
            (
                "The regression analysis health check detects arms "
                "across all trials that are regressing metrics. While metric "
                "regressions can happen (especially in exploratory rounds that use "
                "randomized parameters), users may choose to stop arms that are "
                "regressing company-critical metrics.\n\n"
                "No metric regessions detected."
            ),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import warnings
from contextlib import ExitStack

from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig
from ax.api.utils.instantiation.from_config import (
    _parameter_type_converter,
    parameter_from_config,
)
from ax.api.utils.instantiation.from_struct import experiment_from_struct
from ax.api.utils.structs import ExperimentStruct
from ax.core.experiment import Experiment
from ax.core.formatting_utils import DataType
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType as CoreParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import ParameterConstraint
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase


class TestFromConfig(TestCase):
    def test_create_range_parameter(self) -> None:
        float_config = RangeParameterConfig(
            name="float_param",
            parameter_type="float",
            bounds=(0, 1),
        )

        self.assertEqual(
            parameter_from_config(config=float_config),
            RangeParameter(
                name="float_param",
                parameter_type=CoreParameterType.FLOAT,
                lower=0,
                upper=1,
            ),
        )

        float_config_with_log_scaling = RangeParameterConfig(
            name="float_param_with_log_scaling",
            parameter_type="float",
            bounds=(1e-10, 1),
            scaling="log",
        )

        self.assertEqual(
            parameter_from_config(config=float_config_with_log_scaling),
            RangeParameter(
                name="float_param_with_log_scaling",
                parameter_type=CoreParameterType.FLOAT,
                lower=1e-10,
                upper=1,
                log_scale=True,
            ),
        )

        int_config = RangeParameterConfig(
            name="int_param",
            parameter_type="int",
            bounds=(0, 1),
        )

        self.assertEqual(
            parameter_from_config(config=int_config),
            RangeParameter(
                name="int_param",
                parameter_type=CoreParameterType.INT,
                lower=0,
                upper=1,
            ),
        )

        with warnings.catch_warnings(record=True) as ws:
            step_size_config = RangeParameterConfig(
                name="step_size_param",
                parameter_type="float",
                bounds=(0, 100),
                step_size=10,
            )
        self.assertFalse(any("sort_values" in str(w.message) for w in ws))

        self.assertEqual(
            parameter_from_config(config=step_size_config),
            ChoiceParameter(
                name="step_size_param",
                parameter_type=CoreParameterType.FLOAT,
                values=[
                    0.0,
                    10.0,
                    20.0,
                    30.0,
                    40.0,
                    50.0,
                    60.0,
                    70.0,
                    80.0,
                    90.0,
                    100.0,
                ],
                is_ordered=True,
                sort_values=False,
            ),
        )

        with self.assertRaisesRegex(
            UserInputError,
            "Non-linear parameter scaling is not supported when using step_size",
        ):
            parameter_from_config(
                config=RangeParameterConfig(
                    name="step_size_param_with_scaling",
                    parameter_type="float",
                    bounds=(0, 100),
                    step_size=10,
                    scaling="log",
                )
            )

    def test_create_choice_parameter(self) -> None:
        es = ExitStack()
        ws = es.enter_context(warnings.catch_warnings(record=True))
        choice_config = ChoiceParameterConfig(
            name="choice_param",
            parameter_type="str",
            values=["a", "b", "c"],
        )

        self.assertEqual(
            parameter_from_config(config=choice_config),
            ChoiceParameter(
                name="choice_param",
                parameter_type=CoreParameterType.STRING,
                values=["a", "b", "c"],
            ),
        )

        choice_config_with_order = ChoiceParameterConfig(
            name="choice_param_with_order",
            parameter_type="str",
            values=["a", "b", "c"],
            is_ordered=True,
        )
        self.assertEqual(
            parameter_from_config(config=choice_config_with_order),
            ChoiceParameter(
                name="choice_param_with_order",
                parameter_type=CoreParameterType.STRING,
                values=["a", "b", "c"],
                is_ordered=True,
            ),
        )

        choice_config_with_dependents = ChoiceParameterConfig(
            name="choice_param_with_dependents",
            parameter_type="str",
            values=["a", "b", "c"],
            dependent_parameters={
                "a": ["a1", "a2"],
                "b": ["b1", "b2", "b3"],
            },
        )
        self.assertEqual(
            parameter_from_config(config=choice_config_with_dependents),
            ChoiceParameter(
                name="choice_param_with_dependents",
                parameter_type=CoreParameterType.STRING,
                values=["a", "b", "c"],
                dependents={
                    "a": ["a1", "a2"],
                    "b": ["b1", "b2", "b3"],
                },
            ),
        )

        single_element_choice_config = ChoiceParameterConfig(
            name="single_element_choice_param",
            parameter_type="str",
            values=["a"],
        )
        self.assertEqual(
            parameter_from_config(config=single_element_choice_config),
            FixedParameter(
                name="single_element_choice_param",
                parameter_type=CoreParameterType.STRING,
                value="a",
            ),
        )
        self.assertFalse(any("sort_values" in str(w.message) for w in ws))

    def test_experiment_from_config(self) -> None:
        float_parameter = RangeParameterConfig(
            name="float_param",
            parameter_type="float",
            bounds=(0, 1),
        )
        int_parameter = RangeParameterConfig(
            name="int_param",
            parameter_type="int",
            bounds=(0, 1),
        )
        choice_parameter = ChoiceParameterConfig(
            name="choice_param",
            parameter_type="str",
            values=["a", "b", "c"],
        )

        experiment_struct = ExperimentStruct(
            name="test_experiment",
            parameters=[float_parameter, int_parameter, choice_parameter],
            parameter_constraints=["int_param <= float_param"],
            description="test description",
            experiment_type="TEST",
            owner="miles",
        )

        self.assertEqual(
            experiment_from_struct(struct=experiment_struct),
            Experiment(
                search_space=SearchSpace(
                    parameters=[
                        RangeParameter(
                            name="float_param",
                            parameter_type=CoreParameterType.FLOAT,
                            lower=0,
                            upper=1,
                        ),
                        RangeParameter(
                            name="int_param",
                            parameter_type=CoreParameterType.INT,
                            lower=0,
                            upper=1,
                        ),
                        ChoiceParameter(
                            name="choice_param",
                            parameter_type=CoreParameterType.STRING,
                            values=["a", "b", "c"],
                            is_ordered=False,
                            sort_values=False,
                        ),
                    ],
                    parameter_constraints=[
                        ParameterConstraint(
                            constraint_dict={"int_param": 1, "float_param": -1}, bound=0
                        )
                    ],
                ),
                name="test_experiment",
                description="test description",
                experiment_type="TEST",
                properties={"owners": ["miles"]},
                default_data_type=DataType.MAP_DATA,
            ),
        )

        root_parameter = ChoiceParameterConfig(
            name="root_param",
            parameter_type="str",
            values=["left", "right"],
            dependent_parameters={
                "left": ["float_param"],
                "right": ["int_param"],
            },
        )

        hss_struct = ExperimentStruct(
            parameters=[float_parameter, int_parameter, root_parameter],
            parameter_constraints=["int_param <= float_param"],
            name="test_experiment",
            description="test description",
            experiment_type=None,
            owner="miles",
        )

        self.assertEqual(
            experiment_from_struct(struct=hss_struct),
            Experiment(
                search_space=HierarchicalSearchSpace(
                    parameters=[
                        RangeParameter(
                            name="float_param",
                            parameter_type=CoreParameterType.FLOAT,
                            lower=0,
                            upper=1,
                        ),
                        RangeParameter(
                            name="int_param",
                            parameter_type=CoreParameterType.INT,
                            lower=0,
                            upper=1,
                        ),
                        ChoiceParameter(
                            name="root_param",
                            parameter_type=CoreParameterType.STRING,
                            values=["left", "right"],
                            is_ordered=False,
                            sort_values=False,
                            dependents={
                                "left": ["float_param"],
                                "right": ["int_param"],
                            },
                        ),
                    ],
                    parameter_constraints=[
                        ParameterConstraint(
                            constraint_dict={"int_param": 1, "float_param": -1}, bound=0
                        )
                    ],
                ),
                name="test_experiment",
                description="test description",
                properties={"owners": ["miles"]},
                default_data_type=DataType.MAP_DATA,
            ),
        )

    def test_parameter_type_converter(self) -> None:
        self.assertEqual(
            _parameter_type_converter(parameter_type="bool"),
            CoreParameterType.BOOL,
        )
        self.assertEqual(
            _parameter_type_converter(parameter_type="int"),
            CoreParameterType.INT,
        )
        self.assertEqual(
            _parameter_type_converter(parameter_type="float"),
            CoreParameterType.FLOAT,
        )
        self.assertEqual(
            _parameter_type_converter(parameter_type="str"),
            CoreParameterType.STRING,
        )
        with self.assertRaisesRegex(UserInputError, "Unsupported parameter type"):
            _parameter_type_converter(parameter_type="bad")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from ax.api.utils.instantiation.from_config import parameter_from_config
from ax.api.utils.instantiation.from_string import parse_parameter_constraint
from ax.api.utils.structs import ExperimentStruct
from ax.core.experiment import Experiment
from ax.core.formatting_utils import DataType
from ax.core.parameter_constraint import validate_constraint_parameters
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace


def experiment_from_struct(struct: ExperimentStruct) -> Experiment:
    """Create an Experiment from an ExperimentStruct."""
    parameters = [
        parameter_from_config(config=parameter_config)
        for parameter_config in struct.parameters
    ]

    constraints = [
        parse_parameter_constraint(constraint_str=constraint_str)
        for constraint_str in struct.parameter_constraints
    ]

    # Ensure that all ParameterConstraints are valid and acting on existing parameters
    for constraint in constraints:
        validate_constraint_parameters(
            parameters=[
                parameter
                for parameter in parameters
                if parameter.name in constraint.constraint_dict.keys()
            ]
        )

    if any(p.is_hierarchical for p in parameters):
        search_space = HierarchicalSearchSpace(
            parameters=parameters, parameter_constraints=constraints
        )
    else:
        search_space = SearchSpace(
            parameters=parameters, parameter_constraints=constraints
        )

    return Experiment(
        search_space=search_space,
        name=struct.name,
        description=struct.description,
        experiment_type=struct.experiment_type,
        properties={"owners": [struct.owner]},
        default_data_type=DataType.MAP_DATA,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any

import numpy as np


# pyre-fixme[3]: Return annotation cannot be `Any`.
# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def numpy_type_to_python_type(value: Any) -> Any:
    """If `value` is a Numpy int or float, coerce to a Python int or float.
    This is necessary because some of our transforms return Numpy values.
    """
    if isinstance(value, np.integer):
        value = int(value)  # pragma: nocover (covered by generator tests)
    if isinstance(value, np.floating):
        value = float(value)  # pragma: nocover  (covered by generator tests)
    return value



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from unittest.mock import Mock

import numpy as np
import numpy.typing as npt
from ax.adapter import Adapter
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.registry import Generators
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.relativize import (
    BaseRelativize,
    Relativize,
    relativize,
    RelativizeWithConstantControl,
)
from ax.core import BatchTrial
from ax.core.experiment import Experiment
from ax.core.observation import (
    Observation,
    ObservationData,
    ObservationFeatures,
    observations_from_data,
    recombine_observations,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.types import ComparisonOp
from ax.exceptions.core import DataRequiredError
from ax.generators.base import Generator
from ax.metrics.branin import BraninMetric
from ax.utils.common.testutils import TestCase
from ax.utils.stats.statstools import relativize_data
from ax.utils.testing.core_stubs import (
    get_branin_data_batch,
    get_branin_experiment,
    get_branin_multi_objective_optimization_config,
    get_branin_optimization_config,
    get_branin_with_multi_task,
    get_search_space,
)
from pandas.testing import assert_frame_equal
from pyre_extensions import assert_is_instance, none_throws


class RelativizeDataTest(TestCase):
    relativize_classes: list[type[Transform]] = [
        Relativize,
        RelativizeWithConstantControl,
    ]
    cases: list[tuple[type[Transform], list[tuple[npt.NDArray, npt.NDArray]]]] = [
        (
            Relativize,
            [
                (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                (
                    np.array([-22.56, 98.01652893]),
                    np.array([[604.8, 0.0], [0.0, 512.39669421]]),
                ),
                (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                (np.array([-51.25, 98.4]), np.array([[812.5, 0.0], [0.0, 480.0]])),
            ],
        ),
        (
            RelativizeWithConstantControl,
            [
                (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                (
                    np.array([-20.0, 100.0]),
                    np.array([[400.0, 0.0], [0.0, 115.70247934]]),
                ),
                (np.array([0.0, 0.0]), np.array([[0.0, 0.0], [0.0, 0.0]])),
                (np.array([-50.0, 100.0]), np.array([[750.0, 0.0], [0.0, 160.0]])),
            ],
        ),
    ]

    def test_relativize_transform_requires_a_adapter(self) -> None:
        for relativize_cls in self.relativize_classes:
            with self.assertRaisesRegex(
                AssertionError,
                f"{relativize_cls.__name__} transform requires an adapter",
            ):
                relativize_cls(search_space=None)

    def test_relativize_transform_requires_a_adapter_to_have_status_quo_data(
        self,
    ) -> None:
        for relativize_cls in self.relativize_classes:
            # adapter has no status quo
            sobol = Generators.SOBOL(experiment=get_branin_experiment())
            self.assertIsNone(sobol.status_quo)
            with self.assertRaisesRegex(
                DataRequiredError,
                f"{relativize_cls.__name__} requires status quo data.",
            ):
                relativize_cls(search_space=None, adapter=sobol).transform_observations(
                    observations=[
                        Observation(
                            data=ObservationData(
                                metric_names=["foo"],
                                means=np.array([2]),
                                covariance=np.array([[0.1]]),
                            ),
                            features=ObservationFeatures(parameters={"x": 1}),
                            arm_name="0_0",
                        )
                    ],
                )

            # adapter has status quo
            exp = get_branin_experiment(with_batch=True, with_status_quo=True)
            # making status_quo out of design
            none_throws(exp._status_quo)._parameters["x1"] = 10000.0
            for t in exp.trials.values():
                t.mark_running(no_runner_required=True)
                exp.attach_data(
                    get_branin_data_batch(batch=assert_is_instance(t, BatchTrial))
                )
                t.mark_completed()
            data = exp.fetch_data()
            adapter = Adapter(
                experiment=exp, generator=Generator(), transforms=[relativize_cls]
            )
            mean_in_data = data.df.query(
                f"arm_name == '{none_throws(exp.status_quo).name}'"
            )["mean"].item()
            # adapter.status_quo_data_by_trial is accurate
            self.assertEqual(
                mean_in_data,
                none_throws(adapter.status_quo_data_by_trial)[0].means[0],
            )

            # create a new experiment
            new_exp = get_branin_experiment(
                with_batch=True,
                with_status_quo=True,
            )
            for t in new_exp.trials.values():
                t.mark_running(no_runner_required=True)
                new_exp.attach_data(
                    get_branin_data_batch(batch=assert_is_instance(t, BatchTrial))
                )
                t.mark_completed()
            new_data = new_exp.fetch_data()
            # Construct adapter with the new data.
            adapter = Adapter(
                experiment=new_exp, generator=Generator(), transforms=[relativize_cls]
            )
            # The new data is different from the original data
            self.assertNotEqual(data, new_data)
            self.assertFalse(data.df.equals(new_data.df))
            mean_in_data = new_data.df.query(
                f"arm_name == '{none_throws(new_exp.status_quo).name}'"
            )["mean"].item()
            # adapter.status_quo_data_by_trial remains accurate
            self.assertEqual(
                mean_in_data,
                none_throws(adapter.status_quo_data_by_trial)[0].means[0],
            )
            # test transform edge cases
            observations = observations_from_data(
                experiment=exp,
                data=data,
            )
            tf = relativize_cls(search_space=None, adapter=adapter)
            # making observation coming from trial_index not in adapter
            observations[0].features.trial_index = 999
            self.assertRaises(ValueError, tf.transform_observations, observations)

    def test_relativize_transform_observations(self) -> None:
        def _check_transform_observations(
            tf: Transform,
            observations: list[Observation],
            expected_mean_and_covar: list[tuple[npt.NDArray, npt.NDArray]],
        ) -> None:
            results = tf.transform_observations(observations)
            for i, tsfm_obs in enumerate(results):
                expected_mean, expected_covar = expected_mean_and_covar[i]
                self.assertEqual(tsfm_obs.data.metric_names, metric_names)
                # status quo means must always be zero
                self.assertTrue(
                    np.allclose(tsfm_obs.data.means, expected_mean),
                    tsfm_obs.data.means,
                )
                # status quo covariances must always be zero
                self.assertTrue(
                    np.allclose(tsfm_obs.data.covariance, expected_covar),
                    tsfm_obs.data.covariance,
                )
            # Check untransform
            untsfm_results = tf.untransform_observations(results)
            j = 0
            for untsfm_obs in untsfm_results:
                obs = observations[j]
                # skip status quo for the non-target trial since that
                # is removed when transforming observations
                if untsfm_obs.arm_name != obs.arm_name:
                    j += 1
                    obs = observations[j]
                self.assertTrue(np.allclose(untsfm_obs.data.means, obs.data.means))
                self.assertTrue(
                    np.allclose(untsfm_obs.data.covariance, obs.data.covariance)
                )
                j += 1

        metric_names: list[str] = ["foobar", "foobaz"]
        arm_names = ["status_quo", "0_0", "status_quo", "1_0"]
        obs_data = [
            ObservationData(
                metric_names=metric_names,
                means=np.array([2.5, 5.5]),
                covariance=np.array([[0.2, 0.0], [0.0, 0.3]]),
            ),
            ObservationData(
                metric_names=metric_names,
                means=np.array([2.0, 11.0]),
                covariance=np.array([[0.25, 0.0], [0.0, 0.35]]),
            ),
            ObservationData(
                metric_names=metric_names,
                means=np.array([2.0, 5.0]),
                covariance=np.array([[0.1, 0.0], [0.0, 0.2]]),
            ),
            ObservationData(
                metric_names=metric_names,
                means=np.array([1.0, 10.0]),
                covariance=np.array([[0.3, 0.0], [0.0, 0.4]]),
            ),
        ]
        obs_features = [
            ObservationFeatures(parameters={"x": 1}, trial_index=0),
            ObservationFeatures(parameters={"x": 2}, trial_index=0),
            ObservationFeatures(parameters={"x": 1}, trial_index=1),
            ObservationFeatures(parameters={"x": 3}, trial_index=1),
        ]

        observations = recombine_observations(obs_features, obs_data, arm_names)
        adapter = Mock(
            status_quo=Mock(
                data=obs_data[2], features=obs_features[2], arm_name=arm_names[2]
            ),
            status_quo_name=arm_names[2],
            status_quo_data_by_trial={0: obs_data[0], 1: obs_data[2]},
            _experiment=None,
        )

        for relativize_cls, expected_mean_and_covar in self.cases:
            tf = relativize_cls(
                search_space=None,
                observations=observations,
                adapter=adapter,
            )
            # check transform and untransform on observations
            _check_transform_observations(
                tf=tf,
                observations=observations,
                expected_mean_and_covar=expected_mean_and_covar,
            )
            # transform should still work when trial_index is None
            if relativize_cls in [RelativizeWithConstantControl, Relativize]:
                adapter = Mock(
                    status_quo=Mock(
                        data=obs_data[2],
                        features=obs_features[2],
                        arm_name=arm_names[2],
                    ),
                    status_quo_data_by_trial={0: obs_data[0], 1: obs_data[2]},
                )
                tf = relativize_cls(
                    search_space=None,
                    observations=observations,
                    adapter=adapter,
                )
                observations2 = deepcopy(observations)
                for obs in observations2:
                    obs.features.trial_index = None
                _check_transform_observations(
                    tf=tf,
                    observations=observations2[2:4],
                    expected_mean_and_covar=expected_mean_and_covar[2:4],
                )

    def test_bad_relativize(self) -> None:
        # Check instantiation and subclassing of BaseRelativize
        class BadRelativize(BaseRelativize):
            pass

        for abstract_cls in [BaseRelativize, BadRelativize]:
            with self.assertRaisesRegex(TypeError, "Can't instantiate abstract class"):
                abstract_cls(
                    search_space=None,
                    observations=None,
                    adapter=None,
                )

    def test_transform_status_quos_always_zero(self) -> None:
        for _ in range(1000):
            sq_mean = np.random.uniform(-10.0, 10.0)
            sq_sem = np.random.uniform(0, 10.0)
            mean = np.random.uniform(-10.0, 10.0)
            sem = np.random.uniform(0, 10.0)
            if abs(sq_mean) < 1e-10 or abs(sq_mean) == sq_sem:
                continue

            arm_names = ["status_quo", "0_0"]
            obs_data = [
                ObservationData(
                    metric_names=["foo"],
                    means=np.array([sq_mean]),
                    covariance=np.array([[sq_sem]]),
                ),
                ObservationData(
                    metric_names=["foo"],
                    means=np.array([mean]),
                    covariance=np.array([[sem]]),
                ),
            ]
            obs_features = [
                ObservationFeatures(parameters={"x": 1}, trial_index=0),
                ObservationFeatures(parameters={"x": 2}, trial_index=0),
            ]
            adapter = Mock(
                status_quo=Mock(
                    data=obs_data[0], features=obs_features[0], arm_name=arm_names[0]
                ),
                status_quo_data_by_trial={0: obs_data[0]},
            )
            observations = recombine_observations(obs_features, obs_data, arm_names)
            for relativize_cls in [Relativize, RelativizeWithConstantControl]:
                transform = relativize_cls(search_space=None, adapter=adapter)
                relative_obs = transform.transform_observations(observations)
                self.assertEqual(relative_obs[0].data.metric_names, ["foo"])
                self.assertAlmostEqual(relative_obs[0].data.means[0], 0, places=4)
                self.assertAlmostEqual(
                    relative_obs[0].data.covariance[0][0], 0, places=4
                )

    def test_multitask_data(self) -> None:
        experiment = get_branin_with_multi_task()
        data = experiment.fetch_data()

        observations = observations_from_data(
            experiment=experiment,
            data=data,
        )
        relative_observations = observations_from_data(
            experiment=experiment,
            data=relativize_data(
                data=data,
                status_quo_name="status_quo",
                as_percent=True,
                include_sq=True,
            ),
        )

        sq_obs_data = []
        for i in data.df["trial_index"].unique():
            status_quo_data = data.df.loc[
                (data.df["arm_name"] == "status_quo") & (data.df["trial_index"] == i)
            ]
            sq_obs_data.append(
                ObservationData(
                    metric_names=status_quo_data["metric_name"].to_list(),
                    means=status_quo_data["mean"].to_numpy(),
                    covariance=status_quo_data["sem"].to_numpy()[np.newaxis, :] ** 2,
                )
            )

        adapter = Mock(
            status_quo=Observation(
                data=sq_obs_data[0],
                features=ObservationFeatures(
                    parameters=none_throws(experiment.status_quo).parameters
                ),
                arm_name="status_quo",
            ),
            status_quo_data_by_trial={
                i: sq_obs_data[i] for i in range(len(sq_obs_data))
            },
        )

        # not checking RelativizeWithConstantControl here
        # because relativize_data uses delta method
        transform = Relativize(search_space=None, adapter=adapter)

        relative_obs_t = transform.transform_observations(observations)
        self.maxDiff = None
        # this assertion just checks that order is the same, which
        # is only important for the purposes of this test
        self.assertEqual(
            [datum.data.metric_names for datum in relative_obs_t],
            [datum.data.metric_names for datum in relative_observations],
        )
        means = [
            np.array([datum.data.means for datum in relative_obs_t]),
            np.array([datum.data.means for datum in relative_observations]),
        ]
        # `self.assertAlmostEqual(relative_obs_data, expected_obs_data)`
        # fails 1% of the time, so we check with numpy.
        self.assertTrue(
            all(np.isclose(means[0], means[1])),
            means,
        )
        covariances = [
            np.array([datum.data.covariance for datum in relative_observations]),
            np.array([datum.data.covariance for datum in relative_obs_t]),
        ]
        self.assertTrue(
            all(np.isclose(covariances[0], covariances[1])),
            covariances,
        )

    def test_transform_experiment_data(self) -> None:
        experiment = get_branin_with_multi_task()
        experiment.fetch_data()
        experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )
        adapter = Adapter(experiment=experiment, generator=Generator())
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            t = relativize_cls(search_space=experiment.search_space, adapter=adapter)
            relativized_data = t.transform_experiment_data(
                experiment_data=deepcopy(experiment_data)
            )
            self.assertNotEqual(experiment_data, relativized_data)
            # Check that arm data hasn't changed.
            assert_frame_equal(relativized_data.arm_data, experiment_data.arm_data)
            # Check that observation data was relativized correctly.
            expected_mean, expected_sem = [], []
            for (  # pyre-ignore [23]: Pyre doesn't know about the index structure.
                trial_index,
                arm_name,
            ), row in experiment_data.observation_data.iterrows():
                sq_row = experiment_data.observation_data.loc[
                    (trial_index, "status_quo")
                ]
                if arm_name == "status_quo":
                    mean, sem = 0, 0
                else:
                    mean, sem = relativize(
                        means_t=row["mean", "branin"],
                        sems_t=row["sem", "branin"],
                        mean_c=sq_row["mean", "branin"],
                        sem_c=sq_row["sem", "branin"],
                        as_percent=True,
                        control_as_constant=t.control_as_constant,
                    )
                expected_mean.append(mean)
                expected_sem.append(sem)
            self.assertEqual(
                relativized_data.observation_data[("mean", "branin")].tolist(),
                expected_mean,
            )
            self.assertEqual(
                relativized_data.observation_data[("sem", "branin")].tolist(),
                expected_sem,
            )


class RelativizeDataOptConfigTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        search_space = get_search_space()
        gr = Generators.SOBOL(experiment=Experiment(search_space=search_space)).gen(n=1)
        self.model = Mock(
            search_space=search_space,
            status_quo=Mock(
                features=ObservationFeatures(parameters=gr.arms[0].parameters)
            ),
        )
        self.model.status_quo_data_by_trial = {0: None}

    def test_transform_optimization_config_without_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                adapter=self.model,
            )
            optimization_config = get_branin_optimization_config()
            new_config = relativize.transform_optimization_config(
                optimization_config=optimization_config,
                adapter=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)

    def test_transform_optimization_config_with_relative_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                adapter=self.model,
            )
            optimization_config = get_branin_optimization_config()
            optimization_config.outcome_constraints = [
                OutcomeConstraint(
                    metric=BraninMetric("b2", ["x2", "x1"]),
                    op=ComparisonOp.GEQ,
                    bound=-200.0,
                    relative=True,
                )
            ]
            new_config = relativize.transform_optimization_config(
                optimization_config=deepcopy(optimization_config),
                adapter=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)
            self.assertEqual(
                new_config.outcome_constraints[0].bound,
                optimization_config.outcome_constraints[0].bound,
            )
            self.assertFalse(new_config.outcome_constraints[0].relative)
            # Untransform the constraints
            cons = relativize.untransform_outcome_constraints(
                outcome_constraints=new_config.outcome_constraints,
                fixed_features=Mock(),
            )
            self.assertEqual(cons, optimization_config.outcome_constraints)

    def test_transform_optimization_config_with_non_relative_constraints(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                adapter=self.model,
            )
            optimization_config = get_branin_optimization_config()
            optimization_config.outcome_constraints = [
                OutcomeConstraint(
                    metric=BraninMetric("b2", ["x2", "x1"]),
                    op=ComparisonOp.GEQ,
                    bound=-200.0,
                    relative=False,
                )
            ]
            with self.assertRaisesRegex(ValueError, "All constraints must be relative"):
                relativize.transform_optimization_config(
                    optimization_config=optimization_config,
                    adapter=None,
                    fixed_features=Mock(),
                )

    def test_transform_optimization_config_with_relative_thresholds(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                adapter=self.model,
            )
            optimization_config = get_branin_multi_objective_optimization_config(
                has_objective_thresholds=True,
            )
            for threshold in optimization_config.objective_thresholds:
                threshold.relative = True

            new_config = relativize.transform_optimization_config(
                optimization_config=optimization_config,
                adapter=None,
                fixed_features=Mock(),
            )
            self.assertEqual(new_config.objective, optimization_config.objective)
            self.assertEqual(
                # pyre-fixme[16]: `OptimizationConfig` has no attribute
                #  `objective_thresholds`.
                new_config.objective_thresholds[0].bound,
                optimization_config.objective_thresholds[0].bound,
            )
            self.assertFalse(new_config.objective_thresholds[0].relative)
            self.assertEqual(
                new_config.objective_thresholds[1].bound,
                optimization_config.objective_thresholds[1].bound,
            )
            self.assertFalse(new_config.objective_thresholds[1].relative)

    def test_transform_optimization_config_with_non_relative_thresholds(self) -> None:
        for relativize_cls in [Relativize, RelativizeWithConstantControl]:
            relativize = relativize_cls(
                search_space=None,
                observations=[],
                adapter=self.model,
            )
            optimization_config = get_branin_multi_objective_optimization_config(
                has_objective_thresholds=True,
            )
            optimization_config.objective_thresholds[1].relative = False

            with self.assertRaisesRegex(
                ValueError, "All objective thresholds must be relative"
            ):
                relativize.transform_optimization_config(
                    optimization_config=optimization_config,
                    adapter=None,
                    fixed_features=Mock(),
                )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from random import randint

from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckStatus
from ax.analysis.healthcheck.should_generate_candidates import ShouldGenerateCandidates
from ax.utils.common.testutils import TestCase


class TestShouldGenerateCandidates(TestCase):
    def test_should(self) -> None:
        trial_index = randint(0, 10)
        card = ShouldGenerateCandidates(
            should_generate=True,
            reason="Something reassuring",
            trial_index=trial_index,
        ).compute()
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)
        self.assertEqual(card.subtitle, "Something reassuring")

    def test_should_not(self) -> None:
        trial_index = randint(0, 10)
        card = ShouldGenerateCandidates(
            should_generate=False,
            reason="Something concerning",
            trial_index=trial_index,
        ).compute()
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)
        self.assertEqual(card.subtitle, "Something concerning")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.analysis.plotly.progression import (
    _calculate_wallclock_timeseries,
    ProgressionPlot,
)
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_test_map_data_experiment


class TestProgression(TestCase):
    def test_compute(self) -> None:
        analysis = ProgressionPlot(metric_name="branin_map")

        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()

        experiment = get_test_map_data_experiment(
            num_trials=2, num_fetches=5, num_complete=2
        )

        card = analysis.compute(experiment=experiment)

        self.assertEqual(card.name, "ProgressionPlot")
        self.assertEqual(card.title, "branin_map by progression")
        self.assertEqual(
            card.subtitle,
            (
                "The progression plot tracks the evolution of each metric "
                "over the course of the experiment. This visualization is typically "
                "used to monitor the improvement of metrics over Trial iterations, "
                "but can also be useful in informing decisions about early stopping "
                "for Trials."
            ),
        )
        self.assertEqual(
            {*card.df.columns},
            {"trial_index", "arm_name", "branin_map", "progression", "wallclock_time"},
        )

        self.assertIsNotNone(card.blob)

    def test_calculate_wallclock_timeseries(self) -> None:
        experiment = get_test_map_data_experiment(
            num_trials=2, num_fetches=5, num_complete=2
        )
        wallclock_timeseries = _calculate_wallclock_timeseries(
            experiment=experiment, metric_name="branin_map"
        )

        self.assertEqual(len(wallclock_timeseries), 2)
        self.assertTrue(
            all(len(timeseries) == 5 for timeseries in wallclock_timeseries.values())
        )

        for timeseries in wallclock_timeseries.values():
            self.assertTrue(pd.Series(timeseries).is_monotonic_increasing)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from enum import Enum


class TrialStatus(int, Enum):
    """Enum of trial status.

    General lifecycle of a trial is:::

        CANDIDATE --> STAGED --> RUNNING --> COMPLETED
                  ------------->         --> FAILED (retryable)
                                         --> EARLY_STOPPED (deemed unpromising)
                  -------------------------> ABANDONED (non-retryable)
                  -------------------------> STALE (retryable, never run)

    Trial is marked as a ``CANDIDATE`` immediately upon its creation.

    Trials may be abandoned at any time prior to completion or failure.
    The difference between abandonment and failure is that the ``FAILED`` state
    is meant to express a possibly transient or retryable error, so trials in
    that state may be re-run and arm(s) in them may be resuggested by Ax models
    to be added to new trials.

    ``ABANDONED`` trials on the other end, indicate
    that the trial (and arms(s) in it) should not be rerun or added to new
    trials. A trial might be marked ``ABANDONED`` as a result of human-initiated
    action (if some trial in experiment is poorly-performing, deterministically
    failing etc., and should not be run again in the experiment). It might also
    be marked ``ABANDONED`` in an automated way if the trial's execution
    encounters an error that indicates that the arm(s) in the trial should not
    be evaluated in the experiment again (e.g. the parameterization in a given
    arm deterministically causes trial evaluation to fail). Note that it's also
    possible to abandon a single arm in a `BatchTrial` via
    ``batch.mark_arm_abandoned``.

    ``STALE`` trials represent old candidate trials that were generated but never
    executed. These trials are considered terminal but retryable, meaning they
    can be re-suggested by Ax models since they were never actually run and
    therefore don't provide any data about the performance of their arms.

    Early-stopped refers to trials that were deemed
    unpromising by an early-stopping strategy and therefore terminated.

    Additionally, when trials are deployed, they may be in an intermediate
    staged state (e.g. scheduled but waiting for resources) or immediately
    transition to running. Note that ``STAGED`` trial status is not always
    applicable and depends on the ``Runner`` trials are deployed with
    (and whether a ``Runner`` is present at all; for example, in Ax Service
    API, trials are marked as ``RUNNING`` immediately when generated from
    ``get_next_trial``, skipping the ``STAGED`` status).

    NOTE: Data for abandoned trials (or abandoned arms in batch trials) is
    not passed to the model as part of training data, unless ``fit_abandoned``
    option is specified to adapter. Additionally, data from MapMetrics is
    typically excluded unless the corresponding trial is completed.
    """

    CANDIDATE = 0
    STAGED = 1
    FAILED = 2
    COMPLETED = 3
    RUNNING = 4
    ABANDONED = 5
    DISPATCHED = 6  # Deprecated.
    EARLY_STOPPED = 7
    STALE = 8

    @property
    def is_terminal(self) -> bool:
        """True if trial is completed."""
        return (
            self == TrialStatus.ABANDONED
            or self == TrialStatus.COMPLETED
            or self == TrialStatus.FAILED
            or self == TrialStatus.EARLY_STOPPED
            or self == TrialStatus.STALE
        )

    @property
    def expecting_data(self) -> bool:
        """True if trial is expecting data."""
        return self in STATUSES_EXPECTING_DATA

    @property
    def is_deployed(self) -> bool:
        """True if trial has been deployed but not completed."""
        return self == TrialStatus.STAGED or self == TrialStatus.RUNNING

    @property
    def is_failed(self) -> bool:
        """True if this trial is a failed one."""
        return self == TrialStatus.FAILED

    @property
    def is_abandoned(self) -> bool:
        """True if this trial is an abandoned one."""
        return self == TrialStatus.ABANDONED

    @property
    def is_candidate(self) -> bool:
        """True if this trial is a candidate."""
        return self == TrialStatus.CANDIDATE

    @property
    def is_completed(self) -> bool:
        """True if this trial is a successfully completed one."""
        return self == TrialStatus.COMPLETED

    @property
    def is_running(self) -> bool:
        """True if this trial is a running one."""
        return self == TrialStatus.RUNNING

    @property
    def is_early_stopped(self) -> bool:
        """True if this trial is an early stopped one."""
        return self == TrialStatus.EARLY_STOPPED

    @property
    def is_stale(self) -> bool:
        """True if this trial is a stale one."""
        return self == TrialStatus.STALE

    def __format__(self, fmt: str) -> str:
        """Define `__format__` to avoid pulling the `__format__` from the `int`
        mixin (since its better for statuses to show up as `RUNNING` than as
        just an int that is difficult to interpret).

        E.g. batch trial representation with the overridden method is:
        "BatchTrial(experiment_name='test', index=0, status=TrialStatus.CANDIDATE)".

        Docs on enum formatting: https://docs.python.org/3/library/enum.html#others.
        """
        return f"{self!s}"

    def __repr__(self) -> str:
        return f"{self.__class__}.{self.name}"


DEFAULT_STATUSES_TO_WARM_START: list[TrialStatus] = [
    TrialStatus.RUNNING,
    TrialStatus.COMPLETED,
    TrialStatus.ABANDONED,
    TrialStatus.EARLY_STOPPED,
]

NON_ABANDONED_STATUSES: set[TrialStatus] = set(TrialStatus) - {TrialStatus.ABANDONED}

STATUSES_EXPECTING_DATA: list[TrialStatus] = [
    TrialStatus.RUNNING,
    TrialStatus.COMPLETED,
    TrialStatus.EARLY_STOPPED,
]

FAILED_ABANDONED_CANDIDATE_STATUSES: list[TrialStatus] = [
    TrialStatus.ABANDONED,
    TrialStatus.FAILED,
    TrialStatus.CANDIDATE,
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os
from collections.abc import Mapping
from typing import Any

import torch
from ax.adapter.registry import Cont_X_trans, Generators, Y_trans
from ax.adapter.torch import TorchAdapter
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.benchmark_test_functions.surrogate import SurrogateTestFunction
from ax.benchmark.problems.surrogate.lcbench.utils import (
    BASELINE_VALUES,
    DEFAULT_METRIC_NAME,
    get_lcbench_optimization_config,
)
from ax.core.optimization_config import OptimizationConfig
from ax.exceptions.core import UserInputError
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.kernels import ScaleMaternKernel
from ax.generators.torch.botorch_modular.surrogate import (
    ModelConfig,
    Surrogate,
    SurrogateSpec,
)
from ax.utils.testing.mock import skip_fit_gpytorch_mll_context_manager
from botorch.models import SingleTaskGP
from gpytorch.priors import LogNormalPrior

from pyre_extensions import assert_is_instance


DEFAULT_NUM_TRIALS: int = 30


DEFAULT_AND_OPTIMAL_VALUES: dict[str, tuple[float, float]] = {
    "KDDCup09_appetency": (87.14437173839048, 100.41903197808242),
    "APSFailure": (97.3412499690734, 98.38099041845653),
    "albert": (64.42693765555859, 67.1082934765708),
    "Amazon_employee_access": (80.69975381128579, 98.85943103737361),
    "Australian": (78.15200826093329, 93.0325039665508),
    "Fashion-MNIST": (83.10219231927393, 89.07884250211491),
    "car": (64.26087451215653, 88.77391803474296),
    "christine": (70.9732126619125, 73.29816335805616),
    "cnae-9": (25.89740105397502, 119.55228152861949),
    "covertype": (62.13132918760403, 67.15439170116016),
    "dionis": (11.977294194995338, 101.64303302727558),
    "fabert": (37.72627877151164, 73.30035354875776),
    "helena": (7.455048985077637, 29.78291566900156),
    "higgs": (64.80984463924982, 71.89770865111743),
    "jannis": (58.17868556972097, 62.4080058894946),
    "jasmine": (76.76806487249725, 83.56868288456046),
    "kr-vs-kp": (79.76060013094786, 104.6216855876375),
    "mfeat-factors": (69.85128706899793, 111.67026074027292),
    "nomao": (92.85065022473196, 95.83471144381221),
    "shuttle": (98.86272845879327, 100.37428346365724),
    "sylvine": (83.1596613771663, 98.85179841137813),
    "volkert": (45.361097364985376, 58.133196667029864),
}


def get_lcbench_surrogate() -> Surrogate:
    """Construct a surrogate used to fit the LCBench data.

    Returns:
        A Surrogate with the specification used to fit the LCBench data.
    """
    return Surrogate(
        surrogate_spec=SurrogateSpec(
            model_configs=[
                ModelConfig(
                    botorch_model_class=SingleTaskGP,
                    covar_module_class=ScaleMaternKernel,
                    covar_module_options={
                        "nu": 1.5,
                        "ard_num_dims": 7,
                        "outputscale_prior": LogNormalPrior(-3, 0.0025),
                    },
                    input_transform_classes=None,
                )
            ]
        )
    )


def get_lcbench_benchmark_problem(
    dataset_name: str,
    metric_name: str = DEFAULT_METRIC_NAME,
    num_trials: int = DEFAULT_NUM_TRIALS,
    noise_stds: Mapping[str, float] | float = 0.0,
    observe_noise_stds: bool = False,
) -> BenchmarkProblem:
    """Construct a LCBench benchmark problem.

    Args:
        dataset_name: Must be one of the keys of `DEFAULT_AND_OPTIMAL_VALUES`, which
            correspond to the names of the dataset available in LCBench.
        metric_name: The name of the metric to use for the objective.
        num_trials: The number of optimization trials to run.
        noise_stds: The standard deviation of the observation noise.
        observe_noise_stds: Whether to report the standard deviation of the
            observation noise.

    Returns:
        An LCBench surrogate benchmark problem.
    """

    if dataset_name not in DEFAULT_AND_OPTIMAL_VALUES:
        raise UserInputError(
            f"`dataset_name` must be one of {sorted(DEFAULT_AND_OPTIMAL_VALUES)}"
        )
    _, optimal_value = DEFAULT_AND_OPTIMAL_VALUES[dataset_name]
    base_path = os.path.dirname(os.path.realpath(__file__))
    obj: dict[str, Any] = torch.load(
        f=os.path.join(
            base_path, "transfer_learning_data", f"lcbench_{dataset_name}.pt"
        ),
        weights_only=False,
    )
    optimization_config: OptimizationConfig = get_lcbench_optimization_config(
        metric_name=metric_name,
        observe_noise_sd=observe_noise_stds,
        use_map_metric=False,
    )

    def get_surrogate() -> TorchAdapter:
        """Construct an adapter with the LCBench surrogate and datasets.

        Returns:
            A fitted adapter with the LCBench Surrogate and data.
        """
        # We load the model hyperparameters from the saved state dict.
        with skip_fit_gpytorch_mll_context_manager():
            adapter = Generators.BOTORCH_MODULAR(
                surrogate=get_lcbench_surrogate(),
                experiment=obj["experiment"],
                search_space=obj["experiment"].search_space,
                data=obj["data"],
                transforms=Cont_X_trans + Y_trans,
            )
        assert_is_instance(
            adapter.generator, BoTorchGenerator
        ).surrogate.model.load_state_dict(obj["state_dict"])
        return assert_is_instance(adapter, TorchAdapter)

    name = f"LCBench_Surrogate_{dataset_name}:v1"

    test_function = SurrogateTestFunction(
        name=name, outcome_names=[metric_name], get_surrogate=get_surrogate
    )

    return BenchmarkProblem(
        name=name,
        search_space=obj["experiment"].search_space,
        optimization_config=optimization_config,
        num_trials=num_trials,
        optimal_value=optimal_value,
        baseline_value=BASELINE_VALUES[dataset_name],
        test_function=test_function,
        noise_std=0.0 if noise_stds is None else noise_stds,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable

import numpy as np
import numpy.typing as npt
import torch
from ax.generators.model_utils import tunable_feature_indices
from ax.generators.random.base import RandomGenerator
from ax.generators.types import TConfig
from pyre_extensions import none_throws
from torch.quasirandom import SobolEngine


class SobolGenerator(RandomGenerator):
    """This class specifies the generation algorithm for a Sobol generator.

    As Sobol does not make use of a model, it does not implement
    the fit or predict methods.

    Attributes:
        scramble: If True, permutes the parameter values among
            the elements of the Sobol sequence. Default is True.
        See base `RandomGenerator` for a description of remaining attributes.
    """

    def __init__(
        self,
        deduplicate: bool = True,
        seed: int | None = None,
        init_position: int = 0,
        scramble: bool = True,
        generated_points: npt.NDArray | None = None,
        fallback_to_sample_polytope: bool = False,
    ) -> None:
        super().__init__(
            deduplicate=deduplicate,
            seed=seed,
            init_position=init_position,
            generated_points=generated_points,
            fallback_to_sample_polytope=fallback_to_sample_polytope,
        )
        self.scramble = scramble
        # Initialize engine on gen.
        self._engine: SobolEngine | None = None

    def init_engine(self, n_tunable_features: int) -> SobolEngine:
        """Initialize singleton SobolEngine, only on gen.

        Args:
            n_tunable_features: The number of features which can be
                searched over.

        Returns:
            SobolEngine, which can generate Sobol points.

        """
        if not self._engine:
            self._engine = SobolEngine(
                dimension=n_tunable_features, scramble=self.scramble, seed=self.seed
            ).fast_forward(self.init_position)
        return self._engine

    @property
    def engine(self) -> SobolEngine | None:
        """Return a singleton SobolEngine."""
        return self._engine

    def gen(
        self,
        n: int,
        bounds: list[tuple[float, float]],
        linear_constraints: tuple[npt.NDArray, npt.NDArray] | None = None,
        fixed_features: dict[int, float] | None = None,
        model_gen_options: TConfig | None = None,
        rounding_func: Callable[[npt.NDArray], npt.NDArray] | None = None,
        generated_points: npt.NDArray | None = None,
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """Generate new candidates.

        Args:
            n: Number of candidates to generate.
            bounds: A list of (lower, upper) tuples for each column of X.
            linear_constraints: A tuple of (A, b). For k linear constraints on
                d-dimensional x, A is (k x d) and b is (k x 1) such that
                A x <= b.
            fixed_features: A map {feature_index: value} for features that
                should be fixed to a particular value during generation.
            rounding_func: A function that rounds an optimization result
                appropriately (e.g., according to `round-trip` transformations).
            generated_points: A numpy array of shape `n x d` containing the
                previously generated points to deduplicate against.
        Returns:
            2-element tuple containing

            - (n x d) array of generated points.
            - Uniform weights, an n-array of ones for each point.

        """
        tf_indices = tunable_feature_indices(
            bounds=bounds, fixed_features=fixed_features
        )
        if len(tf_indices) > 0:
            self.init_engine(len(tf_indices))
        points, weights = super().gen(
            n=n,
            bounds=bounds,
            linear_constraints=linear_constraints,
            fixed_features=fixed_features,
            model_gen_options=model_gen_options,
            rounding_func=rounding_func,
            generated_points=generated_points,
        )
        if self.engine:
            self.init_position = none_throws(self.engine).num_generated
        return points, weights

    def _gen_samples(self, n: int, tunable_d: int) -> npt.NDArray:
        """Generate n samples.

        Args:
            n: Number of samples to generate.
            tunable_d: The dimension of the generated samples. This must
                match the tunable parameters used while initializing the
                Sobol engine.

        Returns:
            A numpy array of samples of shape `(n x tunable_d)`.
        """
        if tunable_d == 0:
            return np.zeros((n, 0))
        if self.engine is None:
            raise ValueError(
                "Sobol Engine must be initialized before candidate generation."
            )
        return none_throws(self.engine).draw(n, dtype=torch.double).numpy()



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import math
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
from ax.core.base_trial import BaseTrial
from ax.core.map_data import MAP_KEY, MapData
from ax.core.map_metric import MapMetricFetchResult
from ax.core.metric import MetricFetchE
from ax.metrics.noisy_function_map import NoisyFunctionMapMetric
from ax.utils.common.result import Err, Ok
from ax.utils.measurement.synthetic_functions import branin
from pyre_extensions import assert_is_instance


class BraninTimestampMapMetric(NoisyFunctionMapMetric):
    def __init__(
        self,
        name: str,
        param_names: Iterable[str],
        noise_sd: float = 0.0,
        lower_is_better: bool | None = None,
        rate: float | None = None,
        cache_evaluations: bool = True,
        decay_function_name: str = "exp_decay",
    ) -> None:
        """A Branin map metric with an optional multiplicative factor of
        `1 + decay_function(rate * t)` where `t` is the runtime of the trial.
        `decay_function` is a function of the form `f: float -> float`, either
        an exponential decay or a power law decay, and `rate` is the parameter that
        controls the rate of decay. The functions decay to zero as `t` increases.
        If the multiplicative factor is used, the function is `1 + f(0)` times the usual
        value at `t = 0`, and approaches standard Branin values as `t` grows large.

        Args:
            name: Name of the metric.
            param_names: An ordered list of names of parameters to be passed
                to the deterministic function.
            noise_sd: Scale of normal noise added to the function result.
            lower_is_better: Flag for metrics which should be minimized.
            rate: Parameter of the multiplicative factor.
            decay_function_name: The name of the function used to decay the
                multiplicative factor. Options: "exp_decay", and "pow_decay".
        """
        self.rate = rate
        self.decay_function_name = decay_function_name

        def _exp_decay(x: float) -> float:
            return math.exp(-x)

        def _pow_decay(x: float) -> float:
            return max(x, 1e-3) ** (-1 / np.pi)

        decay_functions = {"exp_decay": _exp_decay, "pow_decay": _pow_decay}

        self._decay_function: Callable[[float], float] = decay_functions[
            decay_function_name
        ]

        # pyre-fixme[4]: Attribute must be annotated.
        self._trial_index_to_timestamp = defaultdict(int)

        super().__init__(
            name=name,
            param_names=param_names,
            noise_sd=noise_sd,
            lower_is_better=lower_is_better,
            cache_evaluations=cache_evaluations,
        )

    def __eq__(self, o: BraninTimestampMapMetric) -> bool:
        """Ignore _timestamp on equality checks"""
        return (
            self.name == o.name
            and self.param_names == o.param_names
            and self.noise_sd == o.noise_sd
            and self.lower_is_better == o.lower_is_better
        )

    def fetch_trial_data(
        self, trial: BaseTrial, noisy: bool = True, **kwargs: Any
    ) -> MapMetricFetchResult:
        try:
            if (
                self._trial_index_to_timestamp[trial.index] == 0
                or trial.status.is_running
            ):
                self._trial_index_to_timestamp[trial.index] += 1

            datas = []
            for timestamp in range(self._trial_index_to_timestamp[trial.index]):
                res = [
                    self.f(
                        np.fromiter(arm.parameters.values(), dtype=float),
                        timestamp=timestamp,
                    )
                    for arm in trial.arms
                ]

                df = pd.DataFrame(
                    {
                        "arm_name": [arm.name for arm in trial.arms],
                        "metric_name": self.name,
                        "sem": self.noise_sd if noisy else 0.0,
                        "trial_index": trial.index,
                        "mean": [
                            item["mean"] + self.noise_sd * np.random.randn()
                            if noisy
                            else 0.0
                            for item in res
                        ],
                        MAP_KEY: [item[MAP_KEY] for item in res],
                    }
                )

                datas.append(MapData(df=df))

            return Ok(value=MapData.from_multiple_map_data(datas))

        except Exception as e:
            return Err(
                MetricFetchE(message=f"Failed to fetch {self.name}", exception=e)
            )

    # pyre-fixme[14]: `f` overrides method defined in `NoisyFunctionMapMetric`
    #  inconsistently.
    def f(self, x: npt.NDArray, timestamp: int) -> Mapping[str, Any]:
        x1, x2 = x

        if self.rate is not None:
            weight = 1.0 + self._decay_function(self.rate * timestamp)
        else:
            weight = 1.0

        mean = assert_is_instance(branin(x1=x1, x2=x2), float) * weight

        return {"mean": mean, MAP_KEY: timestamp}



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Iterable
from logging import Logger

import pandas as pd
from ax.core.experiment import Experiment
from ax.core.map_data import MAP_KEY
from ax.early_stopping.strategies.base import BaseEarlyStoppingStrategy
from ax.exceptions.core import UnsupportedError
from ax.generation_strategy.generation_node import GenerationNode
from ax.utils.common.logger import get_logger

logger: Logger = get_logger(__name__)


class ThresholdEarlyStoppingStrategy(BaseEarlyStoppingStrategy):
    """Implements the strategy of stopping a trial if its performance
    doesn't reach a pre-specified threshold by a certain progression."""

    def __init__(
        self,
        metric_names: Iterable[str] | None = None,
        metric_threshold: float = 0.2,
        min_progression: float | None = 10,
        max_progression: float | None = None,
        min_curves: int | None = 5,
        trial_indices_to_ignore: list[int] | None = None,
        normalize_progressions: bool = False,
    ) -> None:
        """Construct a ThresholdEarlyStoppingStrategy instance.

        Args
            metric_names: A (length-one) list of name of the metric to observe. If
                None will default to the objective metric on the Experiment's
                OptimizationConfig.
            metric_threshold: The metric threshold that a trial needs to reach by
                min_progression in order not to be stopped.
            min_progression: Only stop trials if the latest progression value
                (e.g. timestamp, epochs, training data used) is worse than this
                threshold. Prevents stopping prematurely before enough data is gathered
                to make a decision.
            max_progression: Do not stop trials that have passed `max_progression`.
                Useful if we prefer finishing a trial that are already near completion.
            min_curves: Trials will not be stopped until a number of trials
                `min_curves` have completed with curve data attached. That is, if
                `min_curves` trials are completed but their curve data was not
                successfully retrieved, further trials may not be early-stopped.
            trial_indices_to_ignore: Trial indices that should not be early-stopped.
            normalize_progressions: Normalizes the progression column of the MapData df
                by dividing by the max. If the values were originally in [0, `prog_max`]
                (as we would expect), the transformed values will be in [0, 1]. Useful
                for inferring the max progression and allows `min_progression` to be
                specified in the transformed space. IMPORTANT: Typically, `min_curves`
                should be > 0 to ensure that at least one trial has completed and that
                we have a reliable approximation for `prog_max`.
        """
        super().__init__(
            metric_names=metric_names,
            min_progression=min_progression,
            max_progression=max_progression,
            min_curves=min_curves,
            trial_indices_to_ignore=trial_indices_to_ignore,
            normalize_progressions=normalize_progressions,
        )
        self.metric_threshold = metric_threshold

        if metric_names is not None and len(list(metric_names)) > 1:
            raise UnsupportedError(
                "ThresholdEarlyStoppingStrategy only supports a single metric. Use "
                "LogicalEarlyStoppingStrategy to compose early stopping strategies "
                "with multiple metrics."
            )

    def should_stop_trials_early(
        self,
        trial_indices: set[int],
        experiment: Experiment,
        current_node: GenerationNode | None = None,
    ) -> dict[int, str | None]:
        """Stop a trial if its performance doesn't reach a pre-specified threshold
        by `min_progression`.

        Args:
            trial_indices: Indices of candidate trials to consider for early stopping.
            experiment: Experiment that contains the trials and other contextual data.
            current_node: The current ``GenerationNode`` on the ``GenerationStrategy``
                used to generate trials for the ``Experiment``. Early stopping
                strategies may utilize components of the current node when making
                stopping decisions.

        Returns:
            A dictionary mapping trial indices that should be early stopped to
            (optional) messages with the associated reason. An empty dictionary
            means no suggested updates to any trial's status.
        """
        metric_name, minimize = self._default_objective_and_direction(
            experiment=experiment
        )
        data = self._check_validity_and_get_data(
            experiment=experiment, metric_names=[metric_name]
        )
        if data is None:
            # don't stop any trials if we don't get data back
            return {}

        df = data.map_df

        # default checks on `min_progression` and `min_curves`; if not met, don't do
        # early stopping at all and return {}
        if not self.is_eligible_any(
            trial_indices=trial_indices, experiment=experiment, df=df
        ):
            return {}

        df_objective = df[df["metric_name"] == metric_name]
        decisions = {
            trial_index: self._should_stop_trial_early(
                trial_index=trial_index,
                experiment=experiment,
                df=df_objective,
                minimize=minimize,
            )
            for trial_index in trial_indices
        }
        return {
            trial_index: reason
            for trial_index, (should_stop, reason) in decisions.items()
            if should_stop
        }

    def _should_stop_trial_early(
        self,
        trial_index: int,
        experiment: Experiment,
        df: pd.DataFrame,
        minimize: bool,
    ) -> tuple[bool, str | None]:
        """Stop a trial if its performance doesn't reach a pre-specified threshold
        by `min_progression`.

        Args:
            trial_index: Indices of candidate trial to stop early.
            experiment: Experiment that contains the trials and other contextual data.
            df: Dataframe of partial results for the objective metric.
            minimize: Whether objective value is being minimized.

        Returns:
            A tuple `(should_stop, reason)`, where `should_stop` is `True` iff the
            trial should be stopped, and `reason` is an (optional) string providing
            information on why the trial should or should not be stopped.
        """
        logger.info(f"Considering trial {trial_index} for early stopping.")

        stopping_eligible, reason = self.is_eligible(
            trial_index=trial_index, experiment=experiment, df=df
        )
        if not stopping_eligible:
            return False, reason

        # threshold early stopping logic
        df_trial = df[df["trial_index"] == trial_index].dropna(subset=["mean"])
        trial_last_prog = df_trial[MAP_KEY].max()
        data_last_prog = df_trial[df_trial[MAP_KEY] == trial_last_prog]["mean"].iloc[0]
        logger.info(
            "Early stopping objective at last progression is:\n" f"{data_last_prog}."
        )
        should_early_stop = (
            data_last_prog > self.metric_threshold
            if minimize
            else data_last_prog < self.metric_threshold
        )
        comp = "worse" if should_early_stop else "better"
        reason = (
            f"Trial objective value {data_last_prog} is {comp} than "
            f"the metric threshold {self.metric_threshold:}."
        )
        logger.info(
            f"Early stopping decision for {trial_index}: {should_early_stop}. "
            f"Reason: {reason}"
        )
        return should_early_stop, reason



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from datetime import datetime, timedelta
from typing import Any, Iterable

import pandas as pd

from ax.analysis.healthcheck.metric_fetching_errors import MetricFetchingErrorsAnalysis

from ax.core.base_trial import BaseTrial
from ax.core.data import Data
from ax.core.metric import Metric, MetricFetchE, MetricFetchResult
from ax.generation_strategy.dispatch_utils import choose_generation_strategy_legacy
from ax.service.orchestrator import Orchestrator, OrchestratorOptions
from ax.utils.common.result import Err, Ok
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment

TEST_DATA = Data(
    df=pd.DataFrame(
        [
            {
                "arm_name": "0_0",
                "metric_name": "test_metric",
                "mean": 1.0,
                "sem": 2.0,
                "trial_index": 0,
            }
        ]
    )
)


class TestMetricWithException(Metric):
    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    @classmethod
    def fetch_trial_data_multi(
        cls,
        trial: BaseTrial,
        metrics: Iterable[Metric],
        **kwargs: Any,
    ) -> dict[str, MetricFetchResult]:
        return {
            metric.name: Err(
                MetricFetchE(
                    message="This is what I do",
                    exception=ValueError(
                        "The metric you are fetching is a test metric!"
                    ),
                )
            )
            for metric in metrics
        }


class TestMetricNoException(Metric):
    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    @classmethod
    def fetch_trial_data_multi(
        cls,
        trial: BaseTrial,
        metrics: Iterable[Metric],
        **kwargs: Any,
    ) -> dict[str, MetricFetchResult]:
        return {
            metric.name: Err(MetricFetchE(message="This is what I do", exception=None))
            for metric in metrics
        }


class TestMetricSuccess(Metric):
    @classmethod
    def is_available_while_running(cls) -> bool:
        return True

    @classmethod
    def fetch_trial_data_multi(
        cls,
        trial: BaseTrial,
        metrics: Iterable[Metric],
        **kwargs: Any,
    ) -> dict[str, MetricFetchResult]:
        return {metric.name: Ok(value=TEST_DATA) for metric in metrics}


def create_dummy_traceback_pastes(
    experiment_name: str,
    metric_fetch_errors: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    for e in metric_fetch_errors:
        if e.get("paste") is not None:
            continue

        if e["traceback"]:
            e["paste"] = "P123"
        else:
            e["paste"] = "no traceback"
    return metric_fetch_errors


class TestMetricFetchingErrors(TestCase):
    def test_metric_fetching_errors_with_traceback(self) -> None:
        # GIVEN an experiment with a test metric and running trial
        exp = get_branin_experiment(with_batch=True)
        # it won't fetch an already completed trial
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.add_tracking_metric(TestMetricWithException(name="test_metric"))
        # AND GIVEN that experiment has tried to fetch data through the orchestrator
        orchestrator = Orchestrator(
            experiment=exp,
            generation_strategy=choose_generation_strategy_legacy(
                search_space=exp.search_space
            ),
            options=OrchestratorOptions(),
        )
        orchestrator.poll_and_process_results()
        self.assertEqual(len(exp._metric_fetching_errors), 1)
        # WHEN we compute MetricFetchingErrorsAnalysis with a traceback creator
        card = MetricFetchingErrorsAnalysis(
            add_traceback_paste_callable=create_dummy_traceback_pastes
        ).compute(experiment=exp)
        # THEN we get a card with a dataframe of errors
        self.assertEqual(len(card.df), 1)
        # AND the dataframe has the right columns in the right order
        self.assertEqual(
            list(card.df.columns),
            [
                "trial_index",
                "metric_name",
                "timestamp",
                "reason",
                "traceback",
            ],
        )
        self.assertEqual(
            card.df["trial_index"].iloc[0],
            0,
        )
        self.assertEqual(
            card.df["metric_name"].iloc[0],
            "test_metric",
        )
        self.assertEqual(
            card.df["reason"].iloc[0],
            "Ran into the following exception: ValueError: "
            "The metric you are fetching is a test metric!",
        )
        self.assertEqual(
            card.df["traceback"].iloc[0],
            "P123",
        )
        self.assertLessEqual(
            card.df["timestamp"].iloc[0],
            datetime.now().isoformat(),
        )
        self.assertGreaterEqual(
            card.df["timestamp"].iloc[0],
            (datetime.now() - timedelta(minutes=1)).isoformat(),
        )

    def test_metric_fetching_errors_without_traceback(self) -> None:
        # GIVEN an experiment with a test metric and running trial
        exp = get_branin_experiment(with_batch=True)
        # it won't fetch an already completed trial
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.add_tracking_metric(TestMetricNoException(name="test_metric"))
        # AND GIVEN that experiment has tried to fetch data through the orchestrator
        orchestrator = Orchestrator(
            experiment=exp,
            generation_strategy=choose_generation_strategy_legacy(
                search_space=exp.search_space
            ),
            options=OrchestratorOptions(),
        )
        orchestrator.poll_and_process_results()
        self.assertEqual(len(exp._metric_fetching_errors), 1)
        # WHEN we compute MetricFetchingErrorsAnalysis without a traceback creator
        card = MetricFetchingErrorsAnalysis().compute(experiment=exp)
        # THEN we get a card with a dataframe of errors
        self.assertEqual(len(card.df), 1)
        # AND the dataframe has the right columns in the right order
        self.assertEqual(
            list(card.df.columns),
            [
                "trial_index",
                "metric_name",
                "timestamp",
                "reason",
                "traceback",
            ],
        )
        self.assertEqual(
            card.df["trial_index"].iloc[0],
            0,
        )
        self.assertEqual(
            card.df["metric_name"].iloc[0],
            "test_metric",
        )
        self.assertEqual(card.df["reason"].iloc[0], "This is what I do")
        self.assertEqual(
            card.df["traceback"].iloc[0],
            "No traceback available",
        )
        self.assertLessEqual(
            card.df["timestamp"].iloc[0],
            datetime.now().isoformat(),
        )
        self.assertGreaterEqual(
            card.df["timestamp"].iloc[0],
            (datetime.now() - timedelta(minutes=1)).isoformat(),
        )

    def test_error_order(self) -> None:
        # GIVEN an experiment with a test metric and running trial
        exp = get_branin_experiment(with_batch=True)
        # it won't fetch an already completed trial
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.add_tracking_metric(TestMetricWithException(name="test_metric1"))
        exp.add_tracking_metric(TestMetricWithException(name="test_metric2"))
        # AND GIVEN that experiment has tried to fetch data through the orchestrator
        orchestrator = Orchestrator(
            experiment=exp,
            generation_strategy=choose_generation_strategy_legacy(
                search_space=exp.search_space
            ),
            options=OrchestratorOptions(),
        )
        orchestrator.poll_and_process_results()
        self.assertEqual(len(exp._metric_fetching_errors), 2)
        # WHEN we compute MetricFetchingErrorsAnalysis
        card = MetricFetchingErrorsAnalysis().compute(experiment=exp)
        # THEN we get a cards in descending ts order
        self.assertEqual(len(card.df), 2)
        self.assertGreater(
            card.df["timestamp"].iloc[0],
            card.df["timestamp"].iloc[1],
        )

    def test_error_gets_updated_for_same_metric(self) -> None:
        # This tests that an error in exp._metric_fetching_errors is updated
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.add_tracking_metric(TestMetricWithException(name="test_metric"))

        orchestrator = Orchestrator(
            experiment=exp,
            generation_strategy=choose_generation_strategy_legacy(
                search_space=exp.search_space
            ),
            options=OrchestratorOptions(),
        )
        orchestrator.poll_and_process_results()
        original_ts = exp._metric_fetching_errors[(0, "test_metric")]["timestamp"]
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        orchestrator.poll_and_process_results()

        self.assertEqual(len(exp._metric_fetching_errors), 1)
        card = MetricFetchingErrorsAnalysis().compute(experiment=exp)
        self.assertEqual(len(card.df), 1)
        self.assertGreater(card.df["timestamp"].iloc[0], original_ts)

    def test_error_gets_popped_on_successful_fetch(self) -> None:
        # This tests that an error in exp._metric_fetching_errors is popped
        # on a successful fetch
        exp = get_branin_experiment(with_batch=True)
        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.add_tracking_metric(TestMetricWithException(name="test_metric"))

        orchestrator = Orchestrator(
            experiment=exp,
            generation_strategy=choose_generation_strategy_legacy(
                search_space=exp.search_space
            ),
            options=OrchestratorOptions(),
        )
        orchestrator.poll_and_process_results()
        self.assertEqual(len(exp._metric_fetching_errors), 1)

        exp.trials[0].mark_running(no_runner_required=True, unsafe=True)
        exp.remove_tracking_metric("test_metric")
        exp.add_tracking_metric(TestMetricSuccess(name="test_metric"))
        orchestrator.poll_and_process_results()

        self.assertEqual(len(exp._metric_fetching_errors), 0)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from logging import Logger
from typing import Optional, TYPE_CHECKING

import numpy as np
from ax.adapter.base import unwrap_observation_data
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.ivw import ivw_metric_merge
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import OutcomeConstraint, ScalarizedOutcomeConstraint
from ax.core.types import TModelCov, TModelMean
from ax.exceptions.core import DataRequiredError
from ax.utils.common.logger import get_logger
from pyre_extensions import none_throws


if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


logger: Logger = get_logger(__name__)


class Derelativize(Transform):
    """Changes relative constraints to not-relative constraints using a plug-in
    estimate of the status quo value.

    If status quo is in-design, uses model estimate at status quo. If not, uses
    raw observation at status quo.

    Will raise an error if status quo is in-design and model fails to predict
    for it, unless the flag "use_raw_status_quo" is set to True in the
    transform config, in which case it will fall back to using the observed
    value in the training data.

    Transform is done in-place.
    """

    no_op_for_experiment_data: bool = True

    def transform_optimization_config(
        self,
        optimization_config: OptimizationConfig,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        fixed_features: ObservationFeatures | None = None,
    ) -> OptimizationConfig:
        use_raw_sq = self.config.get("use_raw_status_quo", False)
        has_relative_constraint = any(
            c.relative for c in optimization_config.all_constraints
        )
        if not has_relative_constraint:
            return optimization_config
        # Else, we have at least one relative constraint.
        # Estimate the value at the status quo.
        if adapter is None:
            raise ValueError("Adapter not supplied to transform.")
        # Unobserved status quo corresponds to an adapter.status_quo of None.
        if adapter.status_quo is None:
            raise DataRequiredError(
                "Optimization config has relative constraint, but model was "
                "not fit with status quo."
            )

        sq = none_throws(adapter.status_quo)
        sq_data = ivw_metric_merge(obsd=sq.data, conflicting_noiseless="raise")
        raw_f, _ = unwrap_observation_data([sq_data])
        if not use_raw_sq and adapter.model_space.check_membership(
            sq.features.parameters
        ):
            # Only use model predictions if the status quo is in the search space
            # (including parameter constraints) and `use_raw_sq` is false.
            f, cov = adapter.predict(
                observation_features=[sq.features], use_posterior_predictive=True
            )
            # Warn if the raw SQ values are outside of the CI for the predictions.
            _warn_if_raw_sq_is_out_of_CI(
                optimization_config=optimization_config,
                raw_f=raw_f,
                pred_f=f,
                pred_cov=cov,
            )
        else:
            f = raw_f

        # Plug in the status quo value to each relative constraint.
        for c in optimization_config.all_constraints:
            if c.relative:
                if isinstance(c, ScalarizedOutcomeConstraint):
                    missing_metrics = {
                        metric.name for metric in c.metrics if metric.name not in f
                    }
                    if len(missing_metrics) > 0:
                        raise DataRequiredError(
                            f"Status-quo metric value not yet available for metric(s) "
                            f"{missing_metrics}."
                        )
                    # The sq_val of scalarized outcome is the weighted
                    # sum of its component metrics
                    sq_val = np.sum(
                        [
                            c.weights[i] * f[metric.name][0]
                            for i, metric in enumerate(c.metrics)
                        ]
                    )
                elif c.metric.name in f:
                    sq_val = f[c.metric.name][0]
                else:
                    raise DataRequiredError(
                        f"Status-quo metric value not yet available for metric "
                        f"{c.metric.name}."
                    )
                c.bound = derelativize_bound(bound=c.bound, sq_val=sq_val)
                c.relative = False
        return optimization_config

    def untransform_outcome_constraints(
        self,
        outcome_constraints: list[OutcomeConstraint],
        fixed_features: ObservationFeatures | None = None,
    ) -> list[OutcomeConstraint]:
        # We intentionally leave outcome constraints derelativized when
        # untransforming.
        return outcome_constraints


def _warn_if_raw_sq_is_out_of_CI(
    optimization_config: OptimizationConfig,
    raw_f: TModelMean,
    pred_f: TModelMean,
    pred_cov: TModelCov,
) -> None:
    """Warn if the raw SQ values for relative constraint metrics deviate
    by more than 1.96 standard deviation from the predictions.
    """
    relative_metrics = {
        oc.metric.name
        for oc in optimization_config.all_constraints
        if oc.relative and not isinstance(oc, ScalarizedOutcomeConstraint)
    }.union(
        {
            metric.name
            for oc in optimization_config.all_constraints
            if oc.relative and isinstance(oc, ScalarizedOutcomeConstraint)
            for metric in oc.metrics
        }
    )
    for metric_name in relative_metrics:
        raw_obs = raw_f[metric_name][0]
        pred_mean = pred_f[metric_name][0]
        pred_std = np.sqrt(pred_cov[metric_name][metric_name][0])
        if abs(raw_obs - pred_mean) > 1.96 * pred_std:
            logger.warning(
                "Model predictions for status quo arm for metric "
                f"{metric_name} deviate more than two standard deviations "
                "from the raw status quo value. "
                f"{raw_obs=}, {pred_mean=}, {pred_std=}."
            )


def derelativize_bound(
    bound: float,
    sq_val: float,
) -> float:
    """Derelativize a bound. Note that a positive `bound` makes the derelativized bound
    larger than `sq_val`, i.e. `bound < sq_val`, regardless of the sign of `sq_val`.

    Args:
        bound: The bound to derelativize in percentage terms, so a bound of 1
            corresponds to a 1% increase compared to the status quo.
        sq_val: The status quo value.

    Returns:
        The derelativized bound.

    Examples:
        >>> derelativize_bound(bound=1.0, sq_val=10.0)
        10.1
        >>> derelativize_bound(bound=-1.0, sq_val=10.0)
        9.9
        >>> derelativize_bound(bound=1.0, sq_val=-10.0)
        -9.9
        >>> derelativize_bound(bound=-1.0, sq_val=-10.0)
        -10.1
    """
    return (1 + np.sign(sq_val) * bound / 100.0) * sq_val



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pathlib
from collections.abc import Callable
from typing import Any

import torch
from ax.adapter.base import DataLoaderConfig
from ax.adapter.registry import GeneratorRegistryBase, Generators
from ax.adapter.transforms.base import Transform
from ax.benchmark.benchmark_method import BenchmarkMethod
from ax.benchmark.benchmark_metric import (
    BenchmarkMapMetric,
    BenchmarkMapUnavailableWhileRunningMetric,
    BenchmarkMetric,
    BenchmarkTimeVaryingMetric,
)
from ax.benchmark.benchmark_result import AggregatedBenchmarkResult, BenchmarkResult
from ax.benchmark.benchmark_trial_metadata import BenchmarkTrialMetadata
from ax.core import Experiment, ObservationFeatures
from ax.core.arm import Arm
from ax.core.auxiliary import AuxiliaryExperiment, AuxiliaryExperimentPurpose
from ax.core.batch_trial import AbandonedArm, BatchTrial, GeneratorRunStruct
from ax.core.data import Data
from ax.core.experiment import DataType
from ax.core.generator_run import GeneratorRun
from ax.core.map_data import MapData
from ax.core.map_metric import MapMetric
from ax.core.metric import Metric
from ax.core.multi_type_experiment import MultiTypeExperiment
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
    PreferenceOptimizationConfig,
)
from ax.core.outcome_constraint import ObjectiveThreshold, OutcomeConstraint
from ax.core.parameter import (
    ChoiceParameter,
    DerivedParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    SumConstraint,
)
from ax.core.parameter_distribution import ParameterDistribution
from ax.core.risk_measures import RiskMeasure
from ax.core.search_space import HierarchicalSearchSpace, RobustSearchSpace, SearchSpace
from ax.core.trial import Trial
from ax.core.trial_status import TrialStatus
from ax.core.types import ComparisonOp
from ax.early_stopping.strategies import (
    PercentileEarlyStoppingStrategy,
    ThresholdEarlyStoppingStrategy,
)
from ax.early_stopping.strategies.logical import (
    AndEarlyStoppingStrategy,
    OrEarlyStoppingStrategy,
)
from ax.generation_strategy.best_model_selector import (
    ReductionCriterion,
    SingleDiagnosticBestModelSelector,
)
from ax.generation_strategy.center_generation_node import CenterGenerationNode
from ax.generation_strategy.generation_node import GenerationNode, GenerationStep
from ax.generation_strategy.generation_node_input_constructors import (
    InputConstructorPurpose,
    NodeInputConstructors,
)
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.generation_strategy.transition_criterion import (
    AutoTransitionAfterGen,
    AuxiliaryExperimentCheck,
    IsSingleObjective,
    MaxGenerationParallelism,
    MaxTrials,
    MinimumPreferenceOccurances,
    MinimumTrialsInStatus,
    MinTrials,
    TransitionCriterion,
)
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.surrogate import Surrogate, SurrogateSpec
from ax.generators.torch.botorch_modular.utils import ModelConfig
from ax.generators.winsorization_config import WinsorizationConfig
from ax.global_stopping.strategies.improvement import ImprovementGlobalStoppingStrategy
from ax.metrics.branin import BraninMetric, NegativeBraninMetric
from ax.metrics.branin_map import BraninTimestampMapMetric
from ax.metrics.chemistry import ChemistryMetric, ChemistryProblemType
from ax.metrics.factorial import FactorialMetric
from ax.metrics.hartmann6 import Hartmann6Metric
from ax.metrics.l2norm import L2NormMetric
from ax.metrics.noisy_function import NoisyFunctionMetric
from ax.metrics.sklearn import SklearnDataset, SklearnMetric, SklearnModelType
from ax.runners.synthetic import SyntheticRunner
from ax.service.utils.orchestrator_options import OrchestratorOptions, TrialType
from ax.storage.json_store.decoders import (
    choice_parameter_from_json,
    class_from_json,
    default_from_json,
    fixed_parameter_from_json,
    input_transform_type_from_json,
    multi_objective_from_json,
    outcome_transform_type_from_json,
    pathlib_from_json,
    transform_type_from_json,
)
from ax.storage.json_store.encoders import (
    arm_to_dict,
    auxiliary_experiment_to_dict,
    backend_simulator_to_dict,
    batch_to_dict,
    best_model_selector_to_dict,
    botorch_component_to_dict,
    botorch_model_to_dict,
    botorch_modular_to_dict,
    choice_parameter_to_dict,
    data_to_dict,
    default_to_dict,
    derived_parameter_to_dict,
    experiment_to_dict,
    fixed_parameter_to_dict,
    generation_node_to_dict,
    generation_step_to_dict,
    generation_strategy_to_dict,
    generator_run_to_dict,
    generator_spec_to_dict,
    improvement_global_stopping_strategy_to_dict,
    logical_early_stopping_strategy_to_dict,
    map_data_to_dict,
    metric_to_dict,
    multi_objective_optimization_config_to_dict,
    multi_objective_to_dict,
    multi_type_experiment_to_dict,
    objective_to_dict,
    observation_features_to_dict,
    optimization_config_to_dict,
    order_parameter_constraint_to_dict,
    outcome_constraint_to_dict,
    parameter_constraint_to_dict,
    parameter_distribution_to_dict,
    pathlib_to_dict,
    percentile_early_stopping_strategy_to_dict,
    preference_optimization_config_to_dict,
    range_parameter_to_dict,
    risk_measure_to_dict,
    robust_search_space_to_dict,
    runner_to_dict,
    scalarized_objective_to_dict,
    search_space_to_dict,
    sum_parameter_constraint_to_dict,
    surrogate_to_dict,
    threshold_early_stopping_strategy_to_dict,
    transform_type_to_dict,
    transition_criterion_to_dict,
    trial_to_dict,
    winsorization_config_to_dict,
)
from ax.storage.utils import DomainType, ParameterConstraintType
from ax.utils.common.constants import Keys
from ax.utils.common.serialization import TDecoderRegistry
from ax.utils.testing.backend_simulator import (
    BackendSimulator,
    BackendSimulatorOptions,
    SimTrial,
)
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.models.model import Model
from botorch.models.transforms.input import (
    ChainedInputTransform,
    FilterFeatures,
    InputTransform,
    Normalize,
    Round,
)
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.types import DEFAULT
from gpytorch.constraints import Interval
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.mlls.marginal_log_likelihood import MarginalLogLikelihood
from gpytorch.priors.torch_priors import GammaPrior, LogNormalPrior


# pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
#  avoid runtime subscripting errors.
CORE_ENCODER_REGISTRY: dict[type, Callable[[Any], dict[str, Any]]] = {
    Arm: arm_to_dict,
    AuxiliaryExperiment: auxiliary_experiment_to_dict,
    AndEarlyStoppingStrategy: logical_early_stopping_strategy_to_dict,
    AutoTransitionAfterGen: transition_criterion_to_dict,
    BackendSimulator: backend_simulator_to_dict,
    BatchTrial: batch_to_dict,
    BenchmarkMetric: metric_to_dict,
    BenchmarkMapMetric: metric_to_dict,
    BenchmarkTimeVaryingMetric: metric_to_dict,
    BenchmarkMapUnavailableWhileRunningMetric: metric_to_dict,
    BoTorchGenerator: botorch_model_to_dict,
    BraninMetric: metric_to_dict,
    BraninTimestampMapMetric: metric_to_dict,
    ChainedInputTransform: botorch_component_to_dict,
    ChoiceParameter: choice_parameter_to_dict,
    Data: data_to_dict,
    DerivedParameter: derived_parameter_to_dict,
    Experiment: experiment_to_dict,
    FactorialMetric: metric_to_dict,
    FixedParameter: fixed_parameter_to_dict,
    GammaPrior: botorch_component_to_dict,
    GenerationStep: generation_step_to_dict,
    GenerationNode: generation_node_to_dict,
    GenerationStrategy: generation_strategy_to_dict,
    GeneratorRun: generator_run_to_dict,
    Hartmann6Metric: metric_to_dict,
    ImprovementGlobalStoppingStrategy: improvement_global_stopping_strategy_to_dict,
    Interval: botorch_component_to_dict,
    IsSingleObjective: transition_criterion_to_dict,
    L2NormMetric: metric_to_dict,
    LogNormalPrior: botorch_component_to_dict,
    MapData: map_data_to_dict,
    MapMetric: metric_to_dict,
    MaxGenerationParallelism: transition_criterion_to_dict,
    MaxTrials: transition_criterion_to_dict,
    Metric: metric_to_dict,
    MinTrials: transition_criterion_to_dict,
    MinimumTrialsInStatus: transition_criterion_to_dict,
    MinimumPreferenceOccurances: transition_criterion_to_dict,
    AuxiliaryExperimentCheck: transition_criterion_to_dict,
    GeneratorSpec: generator_spec_to_dict,
    MultiObjective: multi_objective_to_dict,
    MultiObjectiveOptimizationConfig: multi_objective_optimization_config_to_dict,
    MultiTypeExperiment: multi_type_experiment_to_dict,
    Normalize: botorch_component_to_dict,
    FilterFeatures: botorch_component_to_dict,
    PercentileEarlyStoppingStrategy: percentile_early_stopping_strategy_to_dict,
    SklearnMetric: metric_to_dict,
    ChemistryMetric: metric_to_dict,
    NegativeBraninMetric: metric_to_dict,
    NoisyFunctionMetric: metric_to_dict,
    Objective: objective_to_dict,
    ObjectiveThreshold: outcome_constraint_to_dict,
    OptimizationConfig: optimization_config_to_dict,
    OrEarlyStoppingStrategy: logical_early_stopping_strategy_to_dict,
    OrderConstraint: order_parameter_constraint_to_dict,
    OutcomeConstraint: outcome_constraint_to_dict,
    ParameterConstraint: parameter_constraint_to_dict,
    ParameterDistribution: parameter_distribution_to_dict,
    pathlib.Path: pathlib_to_dict,
    pathlib.PurePath: pathlib_to_dict,
    pathlib.PosixPath: pathlib_to_dict,
    pathlib.WindowsPath: pathlib_to_dict,
    pathlib.PurePosixPath: pathlib_to_dict,
    pathlib.PureWindowsPath: pathlib_to_dict,
    PreferenceOptimizationConfig: preference_optimization_config_to_dict,
    RangeParameter: range_parameter_to_dict,
    RiskMeasure: risk_measure_to_dict,
    RobustSearchSpace: robust_search_space_to_dict,
    Round: botorch_component_to_dict,
    TransitionCriterion: transition_criterion_to_dict,
    ScalarizedObjective: scalarized_objective_to_dict,
    SearchSpace: search_space_to_dict,
    SingleDiagnosticBestModelSelector: best_model_selector_to_dict,
    HierarchicalSearchSpace: search_space_to_dict,
    SobolQMCNormalSampler: botorch_component_to_dict,
    SumConstraint: sum_parameter_constraint_to_dict,
    Surrogate: surrogate_to_dict,
    SyntheticRunner: runner_to_dict,
    ThresholdEarlyStoppingStrategy: threshold_early_stopping_strategy_to_dict,
    Trial: trial_to_dict,
    ObservationFeatures: observation_features_to_dict,
    WinsorizationConfig: winsorization_config_to_dict,
}

# Registry for class types, not instances.
# NOTE: Avoid putting a class along with its subclass in `CLASS_ENCODER_REGISTRY`.
# The encoder iterates through this dictionary and uses the first superclass that
# it finds, which might not be the intended superclass.
# pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
#  avoid runtime subscripting errors.
CORE_CLASS_ENCODER_REGISTRY: dict[type, Callable[[Any], dict[str, Any]]] = {
    Acquisition: botorch_modular_to_dict,  # Ax MBM component
    AcquisitionFunction: botorch_modular_to_dict,  # BoTorch component
    InputTransform: botorch_modular_to_dict,  # BoTorch input transform component
    Likelihood: botorch_modular_to_dict,  # BoTorch component
    torch.nn.Module: botorch_modular_to_dict,  # BoTorch module
    MarginalLogLikelihood: botorch_modular_to_dict,  # BoTorch component
    Model: botorch_modular_to_dict,  # BoTorch component
    Transform: transform_type_to_dict,  # Ax general (not just MBM) component
    DEFAULT: default_to_dict,  # BoTorch DEFAULT, used in MBM.
}

# TODO Clean up type signature. Decoders should be allowed to be any method from some
# splattable inputs to the resultant class, not just Types with kwarg inits.
CORE_DECODER_REGISTRY: TDecoderRegistry = {
    "AbandonedArm": AbandonedArm,
    "AndEarlyStoppingStrategy": AndEarlyStoppingStrategy,
    "AutoTransitionAfterGen": AutoTransitionAfterGen,
    "AuxiliaryExperiment": AuxiliaryExperiment,
    "AuxiliaryExperimentCheck": AuxiliaryExperimentCheck,
    "AuxiliaryExperimentPurpose": AuxiliaryExperimentPurpose,
    "Arm": Arm,
    "AggregatedBenchmarkResult": AggregatedBenchmarkResult,
    "BackendSimulator": BackendSimulator,
    "BackendSimulatorOptions": BackendSimulatorOptions,
    "BatchTrial": BatchTrial,
    "BenchmarkMethod": BenchmarkMethod,
    "BenchmarkMetric": BenchmarkMetric,
    "BenchmarkMapMetric": BenchmarkMapMetric,
    "BenchmarkTimeVaryingMetric": BenchmarkTimeVaryingMetric,
    "BenchmarkMapUnavailableWhileRunningMetric": (
        BenchmarkMapUnavailableWhileRunningMetric
    ),
    "BenchmarkResult": BenchmarkResult,
    "BenchmarkTrialMetadata": BenchmarkTrialMetadata,
    "BoTorchGenerator": BoTorchGenerator,
    "BoTorchModel": BoTorchGenerator,
    "BraninMetric": BraninMetric,
    "BraninTimestampMapMetric": BraninTimestampMapMetric,
    "CenterGenerationNode": CenterGenerationNode,
    "ChainedInputTransform": ChainedInputTransform,
    "ChemistryMetric": ChemistryMetric,
    "ChemistryProblemType": ChemistryProblemType,
    "ChoiceParameter": choice_parameter_from_json,
    "ComparisonOp": ComparisonOp,
    "Data": Data,
    "DataLoaderConfig": DataLoaderConfig,
    "DataType": DataType,
    "DerivedParameter": DerivedParameter,
    "DomainType": DomainType,
    "Experiment": Experiment,
    "FactorialMetric": FactorialMetric,
    "FilterFeatures": FilterFeatures,
    "FixedParameter": fixed_parameter_from_json,
    "GammaPrior": GammaPrior,
    "GenerationNode": GenerationNode,
    "GenerationStrategy": GenerationStrategy,
    "GenerationStep": GenerationStep,
    "GeneratorRun": GeneratorRun,
    "GeneratorRunStruct": GeneratorRunStruct,
    "Generators": Generators,
    "GeneratorSpec": GeneratorSpec,
    "Hartmann6Metric": Hartmann6Metric,
    "HierarchicalSearchSpace": HierarchicalSearchSpace,
    "ImprovementGlobalStoppingStrategy": ImprovementGlobalStoppingStrategy,
    "InputConstructorPurpose": InputConstructorPurpose,
    "Interval": Interval,
    "IsSingleObjective": IsSingleObjective,
    "Keys": Keys,
    # DEPRECATED; remains here backward compatibility, with old class
    # name linked to the new corresponding class
    "ListSurrogate": Surrogate,
    "L2NormMetric": L2NormMetric,
    "LogNormalPrior": LogNormalPrior,
    "MapData": MapData,
    "MapMetric": MapMetric,
    "MaxTrials": MaxTrials,
    "MaxGenerationParallelism": MaxGenerationParallelism,
    "Metric": Metric,
    "MinTrials": MinTrials,
    "MinimumTrialsInStatus": MinimumTrialsInStatus,
    "MinimumPreferenceOccurances": MinimumPreferenceOccurances,
    "GeneratorRegistryBase": GeneratorRegistryBase,
    "ModelRegistryBase": GeneratorRegistryBase,
    "ModelConfig": ModelConfig,
    "Models": Generators,
    "ModelSpec": GeneratorSpec,
    "MultiObjective": multi_objective_from_json,
    "MultiObjectiveOptimizationConfig": MultiObjectiveOptimizationConfig,
    "MultiTypeExperiment": MultiTypeExperiment,
    "NegativeBraninMetric": NegativeBraninMetric,
    "NodeInputConstructors": NodeInputConstructors,
    "NoisyFunctionMetric": NoisyFunctionMetric,
    "Normalize": Normalize,
    "Objective": Objective,
    "ObjectiveThreshold": ObjectiveThreshold,
    "OptimizationConfig": OptimizationConfig,
    "OrchestratorOptions": OrchestratorOptions,
    "OrEarlyStoppingStrategy": OrEarlyStoppingStrategy,
    "OrderConstraint": OrderConstraint,
    "OutcomeConstraint": OutcomeConstraint,
    "ParameterConstraint": ParameterConstraint,
    "ParameterConstraintType": ParameterConstraintType,
    "ParameterDistribution": ParameterDistribution,
    "ParameterType": ParameterType,
    "Path": pathlib_from_json,
    "PurePath": pathlib_from_json,
    "PosixPath": pathlib_from_json,
    "WindowsPath": pathlib_from_json,
    "PreferenceOptimizationConfig": PreferenceOptimizationConfig,
    "PurePosixPath": pathlib_from_json,
    "PureWindowsPath": pathlib_from_json,
    "PercentileEarlyStoppingStrategy": PercentileEarlyStoppingStrategy,
    "RangeParameter": RangeParameter,
    "ReductionCriterion": ReductionCriterion,
    "RiskMeasure": RiskMeasure,
    "RobustSearchSpace": RobustSearchSpace,
    "Round": Round,
    "ScalarizedObjective": ScalarizedObjective,
    "SchedulerOptions": OrchestratorOptions,  # DEPRECATED; backward compatibility
    "SearchSpace": SearchSpace,
    "SimTrial": SimTrial,
    "SingleDiagnosticBestModelSelector": SingleDiagnosticBestModelSelector,
    "SklearnDataset": SklearnDataset,
    "SklearnMetric": SklearnMetric,
    "SklearnModelType": SklearnModelType,
    "SumConstraint": SumConstraint,
    "Surrogate": Surrogate,
    "SurrogateMetric": BenchmarkMetric,  # DEPRECATED; backward compatibility
    "SobolQMCNormalSampler": SobolQMCNormalSampler,
    "SyntheticRunner": SyntheticRunner,
    "SurrogateSpec": SurrogateSpec,
    "Trial": Trial,
    "TrialType": TrialType,
    "TrialStatus": TrialStatus,
    "ThresholdEarlyStoppingStrategy": ThresholdEarlyStoppingStrategy,
    "ObservationFeatures": ObservationFeatures,
    "WinsorizationConfig": WinsorizationConfig,
}


# Registry for class types, not instances.
CORE_CLASS_DECODER_REGISTRY: dict[str, Callable[[dict[str, Any]], Any]] = {
    "Type[Acquisition]": class_from_json,
    "Type[AcquisitionFunction]": class_from_json,
    "Type[Kernel]": class_from_json,
    "Type[Likelihood]": class_from_json,
    "Type[torch.nn.Module]": class_from_json,
    "Type[MarginalLogLikelihood]": class_from_json,
    "Type[Model]": class_from_json,
    "Type[Transform]": transform_type_from_json,
    "Type[InputTransform]": input_transform_type_from_json,
    "Type[OutcomeTransform]": outcome_transform_type_from_json,
    "_DefaultType": default_from_json,
}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import itertools
from collections.abc import Callable
from copy import deepcopy
from typing import Any

import numpy.typing as npt
import torch
from ax.adapter.torch import TorchAdapter
from ax.core.search_space import SearchSpaceDigest
from ax.generators.torch.botorch import LegacyBoTorchGenerator
from ax.generators.torch.botorch_modular.generator import (
    BoTorchGenerator as ModularBoTorchGenerator,
)
from ax.utils.sensitivity.derivative_measures import (
    compute_derivatives_from_model_list,
    sample_discrete_parameters,
)
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import Model, ModelList
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.sampling import draw_sobol_samples
from botorch.utils.transforms import unnormalize
from pyre_extensions import assert_is_instance
from torch import Tensor


class SobolSensitivity:
    def __init__(
        self,
        bounds: torch.Tensor,
        input_function: Callable[[torch.Tensor], torch.Tensor] | None = None,
        num_mc_samples: int = 10**4,
        input_qmc: bool = False,
        second_order: bool = False,
        first_order_idcs: torch.Tensor | None = None,
        num_bootstrap_samples: int = 1,
        bootstrap_array: bool = False,
        discrete_features: list[int] | None = None,
    ) -> None:
        r"""Computes three types of Sobol indices:
        first order indices, total indices and second order indices (if specified ).

        Args:
            bounds: Parameter bounds over which to evaluate model sensitivity.
            input_function: The objective function.
            num_mc_samples: The number of montecarlo grid samples
            input_qmc: If True, a qmc Sobol grid is use instead of uniformly random.
            second_order: If True, the second order indices are computed.
            bootstrap: If true, the MC error is returned.
            first_order_idcs: Tensor of previously computed first order indices, where
                first_order_idcs.shape = torch.Size([dim]).
            num_bootstrap_samples: If bootstrap is true, the number of bootstraps has
                to be specified.
            bootstrap_array: If true, all the num_bootstrap_samples extimated indices
                are returned instead of their mean and Var.
            discrete_features: If specified, the inputs associated with the indices in
                this list are generated using an integer-valued uniform distribution,
                rather than the default (pseudo-)random continuous uniform distribution.
        """
        self.input_function = input_function
        self.dim: int = bounds.shape[-1]
        self.num_mc_samples = num_mc_samples
        self.second_order = second_order
        self.bootstrap: bool = num_bootstrap_samples > 1
        self.num_bootstrap_samples: int = (
            num_bootstrap_samples - 1
        )  # deduct 1 because the first is meant to be the full grid
        self.bootstrap_array = bootstrap_array
        self.device: torch.device = bounds.device
        if input_qmc:
            sobol_kwargs = {"bounds": bounds, "n": num_mc_samples, "q": 1}
            seed_A, seed_B = 1234, 5678  # to make it reproducible
            # pyre-ignore
            self.A = draw_sobol_samples(**sobol_kwargs, seed=seed_A).squeeze(1)
            # pyre-ignore
            self.B = draw_sobol_samples(**sobol_kwargs, seed=seed_B).squeeze(1)
        else:
            self.A = unnormalize(
                torch.rand(num_mc_samples, self.dim, device=self.device),
                bounds=bounds,
            )
            self.B = unnormalize(
                torch.rand(num_mc_samples, self.dim, device=self.device),
                bounds=bounds,
            )

        # uniform integral distribution for discrete features
        self.A = sample_discrete_parameters(
            input_mc_samples=self.A,
            discrete_features=discrete_features,
            bounds=bounds,
            num_mc_samples=num_mc_samples,
        )
        self.B = sample_discrete_parameters(
            input_mc_samples=self.B,
            discrete_features=discrete_features,
            bounds=bounds,
            num_mc_samples=num_mc_samples,
        )

        # pyre-fixme[4]: Attribute must be annotated.
        self.A_B_ABi = self.generate_all_input_matrix().to(torch.double)

        if self.bootstrap:
            subset_size = 4
            # pyre-fixme[4]: Attribute must be annotated.
            self.bootstrap_indices = torch.randint(
                0, num_mc_samples, (self.num_bootstrap_samples, subset_size)
            )
        self.f_A: torch.Tensor | None = None
        self.f_B: torch.Tensor | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_ABis: list | None = None
        self.f_total_var: torch.Tensor | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_A_btsp: list | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_B_btsp: list | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_ABis_btsp: list | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_total_var_btsp: list | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_BAis: list | None = None
        # pyre-fixme[24]: Generic type `list` expects 1 type parameter, use
        #  `typing.List` to avoid runtime subscripting errors.
        self.f_BAis_btsp: list | None = None
        self.first_order_idcs: torch.Tensor | None = first_order_idcs
        self.first_order_idcs_btsp: torch.Tensor | None = None

    def generate_all_input_matrix(self) -> torch.Tensor:
        # NOTE Sobol samples of A are ablated with samples of B, and vice versa
        # so baselies are A and B, but each one is ablated by the other for each
        # dimension. First all of A, then (optionally) all of B.
        A_B_ABi_list = [self.A, self.B]
        for i in range(self.dim):
            AB_i = deepcopy(self.A)
            AB_i[:, i] = self.B[:, i]
            A_B_ABi_list.append(AB_i)
        if self.second_order:
            for i in range(self.dim):
                BA_i = deepcopy(self.B)
                BA_i[:, i] = self.A[:, i]
                A_B_ABi_list.append(BA_i)
        A_B_ABi = torch.cat(A_B_ABi_list, dim=0)
        return A_B_ABi

    def evalute_function(self, f_A_B_ABi: torch.Tensor | None = None) -> None:
        r"""evaluates the objective function and devides the evaluation into
            torch.Tensors needed for the indices computation.
        Args:
            f_A_B_ABi: Function evaluations on the entire grid of size M(d+2).
        """
        if f_A_B_ABi is None:
            f_A_B_ABi = self.input_function(self.A_B_ABi)  # pyre-ignore
        # for multiple output models, average the outcomes
        if len(f_A_B_ABi.shape) == 3:
            f_A_B_ABi = f_A_B_ABi.mean(dim=0)

        self.f_A = f_A_B_ABi[: self.num_mc_samples]
        self.f_B = f_A_B_ABi[self.num_mc_samples : self.num_mc_samples * 2]

        # first, there is simply A and B, and then there are all the ablated variants
        # of each (which are retrieved here by slicing the input matrix)
        # but this only retrieves the ablations of A (and not B)
        self.f_ABis = [
            f_A_B_ABi[self.num_mc_samples * (i + 2) : self.num_mc_samples * (i + 3)]
            for i in range(self.dim)
        ]
        # Get the variances of A and B (so simply the variance of 2 num_mc samples)
        self.f_total_var = torch.var(f_A_B_ABi[: self.num_mc_samples * 2])
        if self.bootstrap:
            self.f_A_btsp = [
                torch.index_select(self.f_A, 0, indices)  # pyre-ignore
                for indices in self.bootstrap_indices
            ]
            self.f_B_btsp = [
                torch.index_select(self.f_B, 0, indices)  # pyre-ignore
                for indices in self.bootstrap_indices
            ]
            self.f_ABis_btsp = [
                [
                    torch.index_select(f_ABi, 0, indices)
                    for f_ABi in self.f_ABis  # pyre-ignore
                ]
                for indices in self.bootstrap_indices
            ]
            self.f_total_var_btsp = [
                torch.var(
                    torch.cat(
                        # pyre-fixme[16]: Optional type has no attribute `__getitem__`.
                        (self.f_A_btsp[i], self.f_B_btsp[i]),
                        dim=0,
                    )
                )
                for i in range(self.num_bootstrap_samples)
            ]
        if self.second_order:
            # If second order, we also need to retrieve the ablations of B
            # In total, we have f_ABis and f_BAis which are both of size M(d), and are
            # the ablations of each other
            self.f_BAis = [
                f_A_B_ABi[
                    self.num_mc_samples * (i + 2 + self.dim) : self.num_mc_samples
                    * (i + 3 + self.dim)
                ]
                for i in range(self.dim)
            ]
            if self.bootstrap:
                self.f_BAis_btsp = [
                    [torch.index_select(f_BAi, 0, indices) for f_BAi in self.f_BAis]
                    for indices in self.bootstrap_indices
                ]

    def first_order_indices(self) -> Tensor:
        r"""Computes the first order Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim
            else
                Tensor: (values)x dim
        """
        first_order_idcs = []
        for i in range(self.dim):
            # The only difference between self.f_ABis[i] and self.f_A is in dimension i
            # So we get the variance in the component that corresponds to dimension i
            vi = (
                torch.mean(self.f_B * (self.f_ABis[i] - self.f_A))  # pyre-ignore
                # pyre-fixme[58]: `/` is not supported for operand types `Tensor`
                #  and `Optional[Tensor]`.
                / self.f_total_var
            )
            first_order_idcs.append(vi.unsqueeze(0))
        self.first_order_idcs = torch.cat(first_order_idcs, dim=0).detach()
        if not self.bootstrap:
            return self.first_order_idcs
        else:
            first_order_idcs_btsp = [torch.cat(first_order_idcs, dim=0).unsqueeze(0)]
            for b in range(self.num_bootstrap_samples):
                first_order_idcs = []
                for i in range(self.dim):
                    vi = (
                        torch.mean(
                            self.f_B_btsp[b]
                            * (self.f_ABis_btsp[b][i] - self.f_A_btsp[b])
                        )
                        / self.f_total_var_btsp[b]
                    )
                    first_order_idcs.append(vi.unsqueeze(0))
                first_order_idcs_btsp.append(
                    torch.cat(first_order_idcs, dim=0).unsqueeze(0)
                )
            self.first_order_idcs_btsp = torch.cat(first_order_idcs_btsp, dim=0)
            if self.bootstrap_array:
                return self.first_order_idcs_btsp.detach()
            else:
                return (
                    torch.cat(
                        [
                            self.first_order_idcs_btsp.mean(dim=0).unsqueeze(0),
                            self.first_order_idcs_btsp.var(  # pyre-ignore
                                dim=0
                            ).unsqueeze(0),
                            torch.sqrt(
                                self.first_order_idcs_btsp.var(dim=0)
                                / (self.num_bootstrap_samples + 1)
                            ).unsqueeze(0),
                        ],
                        dim=0,
                    )
                    .t()
                    .detach()
                )

    def total_order_indices(self) -> Tensor:
        r"""Computes the total Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim
            else
                Tensor: (values)x dim
        """
        total_order_idcs = []
        for i in range(self.dim):
            vti = (
                0.5
                # pyre-fixme[58]: `-` is not supported for operand types
                #  `Optional[torch._tensor.Tensor]` and `Any`.
                # pyre-fixme[16]: `Optional` has no attribute `__getitem__`.
                * torch.mean(torch.pow(self.f_A - self.f_ABis[i], 2))
                # pyre-fixme[58]: `/` is not supported for operand types `Tensor`
                #  and `Optional[Tensor]`.
                / self.f_total_var
            )
            total_order_idcs.append(vti.unsqueeze(0))
        if not (self.bootstrap):
            total_order_idcs = torch.cat(total_order_idcs, dim=0).detach()
            return total_order_idcs
        else:
            total_order_idcs_btsp = [torch.cat(total_order_idcs, dim=0).unsqueeze(0)]
            for b in range(self.num_bootstrap_samples):
                total_order_idcs = []
                for i in range(self.dim):
                    vti = (
                        0.5
                        * torch.mean(
                            torch.pow(self.f_A_btsp[b] - self.f_ABis_btsp[b][i], 2)
                        )
                        / self.f_total_var_btsp[b]
                    )
                    total_order_idcs.append(vti.unsqueeze(0))
                total_order_idcs_btsp.append(
                    torch.cat(total_order_idcs, dim=0).unsqueeze(0)
                )
            total_order_idcs_btsp = torch.cat(total_order_idcs_btsp, dim=0)
            if self.bootstrap_array:
                return total_order_idcs_btsp.detach()
            else:
                return (
                    torch.cat(
                        [
                            total_order_idcs_btsp.mean(dim=0).unsqueeze(0),
                            total_order_idcs_btsp.var(dim=0).unsqueeze(0),
                            torch.sqrt(
                                total_order_idcs_btsp.var(dim=0)
                                / (self.num_bootstrap_samples + 1)
                            ).unsqueeze(0),
                        ],
                        dim=0,
                    )
                    .t()
                    .detach()
                )

    def second_order_indices(
        self,
        first_order_idcs: torch.Tensor | None = None,
        first_order_idcs_btsp: torch.Tensor | None = None,
    ) -> Tensor:
        r"""Computes the Second order Sobol indices:
        Args:
            first_order_idcs: Tensor of previously computed first order indices, where
                first_order_idcs.shape = torch.Size([dim]).
            first_order_idcs_btsp: Tensor of all first order indices given by bootstrap.
        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim
            else
                Tensor: (values)x dim
        """
        if not self.second_order:
            raise ValueError(
                "Second order indices has to be specified in the sensitivity definition"
            )

        # TODO Improve this part of the code by vectorization T204291129
        if first_order_idcs is None:
            if self.first_order_idcs is None:
                self.first_order_indices()
            first_order_idcs = self.first_order_idcs
        second_order_idcs = []
        for i in range(self.dim):
            for j in range(i + 1, self.dim):
                vij = torch.mean(
                    self.f_BAis[i] * self.f_ABis[j] - self.f_A * self.f_B  # pyre-ignore
                )
                vij = (
                    # pyre-fixme[58]: `/` is not supported for operand types
                    #  `Tensor` and `Optional[Tensor]`.
                    (vij / self.f_total_var)
                    - first_order_idcs[i]  # pyre-ignore
                    - first_order_idcs[j]
                )
                second_order_idcs.append(vij.unsqueeze(0))
        if not self.bootstrap:
            second_order_idcs = torch.cat(second_order_idcs, dim=0).detach()
            return second_order_idcs
        else:
            second_order_idcs_btsp = [torch.cat(second_order_idcs, dim=0).unsqueeze(0)]
            if first_order_idcs_btsp is None:
                first_order_idcs_btsp = self.first_order_idcs_btsp
            for b in range(self.num_bootstrap_samples):
                second_order_idcs = []
                for i in range(self.dim):
                    for j in range(i + 1, self.dim):
                        vij = torch.mean(
                            self.f_BAis_btsp[b][i] * self.f_ABis_btsp[b][j]
                            - self.f_A_btsp[b] * self.f_B_btsp[b]
                        )
                        vij = (
                            (vij / self.f_total_var_btsp[b])
                            - first_order_idcs_btsp[b][i]
                            - first_order_idcs_btsp[b][j]
                        )
                        second_order_idcs.append(vij.unsqueeze(0))
                second_order_idcs_btsp.append(
                    torch.cat(second_order_idcs, dim=0).unsqueeze(0)
                )
            second_order_idcs_btsp = torch.cat(second_order_idcs_btsp, dim=0)
            if self.bootstrap_array:
                return second_order_idcs_btsp.detach()
            else:
                return (
                    torch.cat(
                        [
                            second_order_idcs_btsp.mean(dim=0).unsqueeze(0),
                            second_order_idcs_btsp.var(dim=0).unsqueeze(0),
                            torch.sqrt(
                                second_order_idcs_btsp.var(dim=0)
                                / (self.num_bootstrap_samples + 1)
                            ).unsqueeze(0),
                        ],
                        dim=0,
                    )
                    .t()
                    .detach()
                )


def GaussianLinkMean(mean: torch.Tensor, var: torch.Tensor) -> torch.Tensor:
    return mean


def ProbitLinkMean(mean: torch.Tensor, var: torch.Tensor) -> torch.Tensor:
    a = mean / torch.sqrt(1 + var)
    return torch.distributions.Normal(0, 1).cdf(a)


class SobolSensitivityGPMean:
    def __init__(
        self,
        model: GPyTorchModel,
        bounds: torch.Tensor,
        num_mc_samples: int = 10**4,
        second_order: bool = False,
        input_qmc: bool = False,
        num_bootstrap_samples: int = 1,
        link_function: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = GaussianLinkMean,
        discrete_features: list[int] | None = None,
    ) -> None:
        r"""Computes three types of Sobol indices:
        first order indices, total indices and second order indices (if specified ).

        Args:
            model: BoTorch model whose posterior is a `GPyTorchPosterior`.
            bounds: `2 x d` parameter bounds over which to evaluate model sensitivity.
            method: if "predictive mean", the predictive mean is used for indices
                computation. If "GP samples", posterior sampling is used instead.
            num_mc_samples: The number of montecarlo grid samples
            second_order: If True, the second order indices are computed.
            input_qmc: If True, a qmc Sobol grid is use instead of uniformly random.
            num_bootstrap_samples: If bootstrap is true, the number of bootstraps has
                to be specified.
            link_function: The link function to be used when computing the indices.
                Indices can be computed for the mean or on samples of the posterior,
                predictive, but defaults to computing on the mean (GaussianLinkMean).
            discrete_features: If specified, the inputs associated with the indices in
                this list are generated using an integer-valued uniform distribution,
                rather than the default (pseudo-)random continuous uniform distribution.
        """
        self.model = model
        self.second_order = second_order
        self.input_qmc = input_qmc
        self.bootstrap: bool = num_bootstrap_samples > 1
        self.num_bootstrap_samples = num_bootstrap_samples
        self.num_mc_samples = num_mc_samples

        def input_function(x: Tensor) -> Tensor:
            assert x.ndim == 2, "Input must be a 2D tensor."
            with torch.no_grad():
                # NOTE: Do not unsqueeze x_ here. Likely due to a matmul issue
                # this ends up using a lot of memory with batched models.
                # See https://github.com/pytorch/botorch/issues/2310.
                means, variances = [], []
                for x_ in x.split(4096):
                    p = self.model.posterior(x_)
                    means.append(p.mean)
                    variances.append(p.variance)
                mean = torch.cat(means, dim=-2)
                variance = torch.cat(variances, dim=-2)
            return link_function(mean, variance)

        self.sensitivity = SobolSensitivity(
            bounds=bounds,
            num_mc_samples=self.num_mc_samples,
            input_function=input_function,
            second_order=self.second_order,
            input_qmc=self.input_qmc,
            num_bootstrap_samples=self.num_bootstrap_samples,
            discrete_features=discrete_features,
        )
        self.sensitivity.evalute_function()

    def first_order_indices(self) -> Tensor:
        r"""Computes the first order Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim
            else
                Tensor: (values)x dim
        """
        return self.sensitivity.first_order_indices()

    def total_order_indices(self) -> Tensor:
        r"""Computes the total Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim
            else
                Tensor: (values)x dim
        """
        return self.sensitivity.total_order_indices()

    def second_order_indices(self) -> Tensor:
        r"""Computes the Second order Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values,var_mc,stderr_mc)x dim(dim-1)/2
            else
                Tensor: (values)x dim(dim-1)/2
        """
        return self.sensitivity.second_order_indices()


class SobolSensitivityGPSampling:
    def __init__(
        self,
        model: Model,
        bounds: torch.Tensor,
        num_gp_samples: int = 10**3,
        num_mc_samples: int = 10**4,
        second_order: bool = False,
        input_qmc: bool = False,
        gp_sample_qmc: bool = False,
        num_bootstrap_samples: int = 1,
        discrete_features: list[int] | None = None,
    ) -> None:
        r"""Computes three types of Sobol indices:
        first order indices, total indices and second order indices (if specified ).

        Args:
            model: Botorch model.
            bounds: `2 x d` parameter bounds over which to evaluate model sensitivity.
            num_gp_samples: If method is "GP samples", the number of GP samples
                has to be set.
            num_mc_samples: The number of montecarlo grid samples
            second_order: If True, the second order indices are computed.
            input_qmc: If True, a qmc Sobol grid is use instead of uniformly random.
            gp_sample_qmc: If True, the posterior sampling is done using
                SobolQMCNormalSampler.
            num_bootstrap_samples: If bootstrap is true, the number of bootstraps has
                to be specified.
            discrete_features: If specified, the inputs associated with the indices in
                this list are generated using an integer-valued uniform distribution,
                rather than the default (pseudo-)random continuous uniform distribution.
        """
        self.model = model
        self.second_order = second_order
        self.input_qmc = input_qmc
        self.gp_sample_qmc = gp_sample_qmc
        # pyre-fixme[4]: Attribute must be annotated.
        self.bootstrap = num_bootstrap_samples > 1
        self.num_bootstrap_samples = num_bootstrap_samples
        self.num_mc_samples = num_mc_samples
        self.num_gp_samples = num_gp_samples
        self.sensitivity = SobolSensitivity(
            bounds=bounds,
            num_mc_samples=self.num_mc_samples,
            second_order=self.second_order,
            input_qmc=self.input_qmc,
            num_bootstrap_samples=self.num_bootstrap_samples,
            bootstrap_array=True,
            discrete_features=discrete_features,
        )
        # TODO: Ideally, we would reduce the memory consumption here as well
        # but this is a tricky since it uses joint posterior sampling.
        posterior = self.model.posterior(self.sensitivity.A_B_ABi)
        if self.gp_sample_qmc:
            sampler = SobolQMCNormalSampler(
                sample_shape=torch.Size([self.num_gp_samples]), seed=0
            )
            # pyre-fixme[4]: Attribute must be annotated.
            self.samples = sampler(posterior)
        else:
            with torch.no_grad():
                self.samples = posterior.rsample(torch.Size([self.num_gp_samples]))

    @property
    def dim(self) -> int:
        """Returns the input dimensionality of `self.model`."""
        return self.sensitivity.dim

    def first_order_indices(self) -> Tensor:
        r"""Computes the first order Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values, var_gp, stderr_gp, var_mc, stderr_mc) x dim
            else
                Tensor: (values, var, stderr) x dim
        """
        first_order_idcs_list = []
        for j in range(self.num_gp_samples):
            self.sensitivity.evalute_function(self.samples[j])
            first_order_idcs = self.sensitivity.first_order_indices()
            first_order_idcs_list.append(first_order_idcs.unsqueeze(0))
        # pyre-fixme[16]: `SobolSensitivityGPSampling` has no attribute
        #  `first_order_idcs_list`.
        self.first_order_idcs_list = torch.cat(first_order_idcs_list, dim=0)
        if not (self.bootstrap):
            first_order_idcs_mean_var_se = []
            for i in range(self.dim):
                first_order_idcs_mean_var_se.append(
                    torch.tensor(
                        [
                            torch.mean(self.first_order_idcs_list[:, i]),
                            torch.var(self.first_order_idcs_list[:, i]),
                            torch.sqrt(
                                torch.var(self.first_order_idcs_list[:, i])
                                / self.num_gp_samples
                            ),
                        ]
                    ).unsqueeze(0)
                )
            first_order_idcs_mean_var_se = torch.cat(
                first_order_idcs_mean_var_se, dim=0
            )
            return first_order_idcs_mean_var_se
        else:
            var_per_bootstrap = torch.var(self.first_order_idcs_list, dim=0)
            gp_var = torch.mean(var_per_bootstrap, dim=0)
            gp_se = torch.sqrt(gp_var / self.num_gp_samples)
            var_per_gp_sample = torch.var(self.first_order_idcs_list, dim=1)
            mc_var = torch.mean(var_per_gp_sample, dim=0)
            mc_se = torch.sqrt(mc_var / (self.num_bootstrap_samples + 1))
            total_mean = self.first_order_idcs_list.reshape(-1, self.dim).mean(dim=0)
            first_order_idcs_mean_vargp_segp_varmc_segp = torch.cat(
                [
                    torch.tensor(
                        [total_mean[i], gp_var[i], gp_se[i], mc_var[i], mc_se[i]]
                    ).unsqueeze(0)
                    for i in range(self.dim)
                ],
                dim=0,
            )
            return first_order_idcs_mean_vargp_segp_varmc_segp

    def total_order_indices(self) -> Tensor:
        r"""Computes the total Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values, var_gp, stderr_gp, var_mc, stderr_mc) x dim
            else
                Tensor: (values, var, stderr) x dim
        """
        total_order_idcs_list = []
        for j in range(self.num_gp_samples):
            self.sensitivity.evalute_function(self.samples[j])
            total_order_idcs = self.sensitivity.total_order_indices()
            total_order_idcs_list.append(total_order_idcs.unsqueeze(0))
        total_order_idcs_list = torch.cat(total_order_idcs_list, dim=0)
        if not (self.bootstrap):
            total_order_idcs_mean_var = []
            for i in range(self.dim):
                total_order_idcs_mean_var.append(
                    torch.tensor(
                        [
                            torch.mean(total_order_idcs_list[:, i]),
                            torch.var(total_order_idcs_list[:, i]),
                            torch.sqrt(
                                torch.var(total_order_idcs_list[:, i])
                                / self.num_gp_samples
                            ),
                        ]
                    ).unsqueeze(0)
                )
            total_order_idcs_mean_var = torch.cat(total_order_idcs_mean_var, dim=0)
            return total_order_idcs_mean_var
        else:
            var_per_bootstrap = torch.var(total_order_idcs_list, dim=0)
            gp_var = torch.mean(var_per_bootstrap, dim=0)
            gp_se = torch.sqrt(gp_var / self.num_gp_samples)
            var_per_gp_sample = torch.var(total_order_idcs_list, dim=1)
            mc_var = torch.mean(var_per_gp_sample, dim=0)
            mc_se = torch.sqrt(mc_var / (self.num_bootstrap_samples + 1))
            total_mean = total_order_idcs_list.reshape(-1, self.dim).mean(dim=0)
            total_order_idcs_mean_vargp_segp_varmc_segp = torch.cat(
                [
                    torch.tensor(
                        [total_mean[i], gp_var[i], gp_se[i], mc_var[i], mc_se[i]]
                    ).unsqueeze(0)
                    for i in range(self.dim)
                ],
                dim=0,
            )
            return total_order_idcs_mean_vargp_segp_varmc_segp

    def second_order_indices(self) -> Tensor:
        r"""Computes the Second order Sobol indices:

        Returns:
            if num_bootstrap_samples>1
                Tensor: (values, var_gp, stderr_gp, var_mc, stderr_mc) x dim(dim-1) / 2
            else
                Tensor: (values, var, stderr) x dim(dim-1) / 2
        """
        if not (self.bootstrap):
            second_order_idcs_list = []
            for j in range(self.num_gp_samples):
                self.sensitivity.evalute_function(self.samples[j])
                second_order_idcs = self.sensitivity.second_order_indices(
                    # pyre-fixme[16]: `SobolSensitivityGPSampling` has no attribute
                    #  `first_order_idcs_list`.
                    self.first_order_idcs_list[j]
                )
                second_order_idcs_list.append(second_order_idcs.unsqueeze(0))
            second_order_idcs_list = torch.cat(second_order_idcs_list, dim=0)
            second_order_idcs_mean_var = []
            # pyre-fixme[61]: `second_order_idcs` is undefined, or not always defined.
            for i in range(len(second_order_idcs)):
                second_order_idcs_mean_var.append(
                    torch.tensor(
                        [
                            torch.mean(second_order_idcs_list[:, i]),
                            torch.var(second_order_idcs_list[:, i]),
                            torch.sqrt(
                                torch.var(second_order_idcs_list[:, i])
                                / self.num_gp_samples
                            ),
                        ]
                    ).unsqueeze(0)
                )
            second_order_idcs_mean_var = torch.cat(second_order_idcs_mean_var, dim=0)
            return second_order_idcs_mean_var
        else:
            second_order_idcs_list = []
            for j in range(self.num_gp_samples):
                self.sensitivity.evalute_function(self.samples[j])
                second_order_idcs = self.sensitivity.second_order_indices(
                    self.first_order_idcs_list[j][0],
                    self.first_order_idcs_list[j][1:],
                )
                second_order_idcs_list.append(second_order_idcs.unsqueeze(0))
            second_order_idcs_list = torch.cat(second_order_idcs_list, dim=0)
            var_per_bootstrap = torch.var(second_order_idcs_list, dim=0)
            gp_var = torch.mean(var_per_bootstrap, dim=0)
            gp_se = torch.sqrt(gp_var / self.num_gp_samples)
            var_per_gp_sample = torch.var(second_order_idcs_list, dim=1)
            mc_var = torch.mean(var_per_gp_sample, dim=0)
            mc_se = torch.sqrt(mc_var / (self.num_bootstrap_samples + 1))
            num_second_order = second_order_idcs_list.shape[-1]
            total_mean = second_order_idcs_list.reshape(-1, num_second_order).mean(
                dim=0
            )
            second_order_idcs_mean_vargp_segp_varmc_segp = torch.cat(
                [
                    torch.tensor(
                        [total_mean[i], gp_var[i], gp_se[i], mc_var[i], mc_se[i]]
                    ).unsqueeze(0)
                    for i in range(num_second_order)
                ],
                dim=0,
            )
            return second_order_idcs_mean_vargp_segp_varmc_segp


def compute_sobol_indices_from_model_list(
    model_list: list[GPyTorchModel],
    bounds: Tensor,
    order: str = "first",
    discrete_features: list[int] | None = None,
    **sobol_kwargs: Any,
) -> Tensor:
    """
    Computes Sobol indices of a list of models on a bounded domain.

    Args:
        model_list: A list of botorch.models.model.Model types for which to compute
            the Sobol indices.
        bounds: A 2 x d Tensor of lower and upper bounds of the domain of the models.
        order: A string specifying the order of the Sobol indices to be computed.
            Supports "first", "second" and "total" and defaults to "first". "total"
            computes the importance of a variable considering its main effect and
            all of its higher-order interactions, whereas "first" and "second"
            the variance when altering the variable in isolation or with one other
            variable, respectively.
        discrete_features: If specified, the inputs associated with the indices in
            this list are generated using an integer-valued uniform distribution,
            rather than the default (pseudo-)random continuous uniform distribution.
        sobol_kwargs: keyword arguments passed on to SobolSensitivityGPMean.

    Returns:
        With m GPs, returns a (m x d) tensor of `order`-order Sobol indices.
    """
    if order not in ["first", "total", "second"]:
        raise NotImplementedError(
            f"Order {order} is not supported. Plese choose one of"
            " 'first', 'total' or 'second'."
        )
    indices = []
    method = getattr(SobolSensitivityGPMean, f"{order}_order_indices")
    second_order = order == "second"
    for model in model_list:
        sens_class = SobolSensitivityGPMean(
            model=model,
            bounds=bounds,
            discrete_features=discrete_features,
            second_order=second_order,
            **sobol_kwargs,
        )
        indices.append(method(sens_class))
    return torch.stack(indices)


def ax_parameter_sens(
    adapter: TorchAdapter,
    metrics: list[str] | None = None,
    order: str = "first",
    signed: bool = True,
    **sobol_kwargs: Any,
) -> dict[str, dict[str, npt.NDArray]]:
    """
    Compute sensitivity for all metrics on an TorchAdapter.

    Sobol measures are always positive regardless of the direction in which the
    parameter influences f. If `signed` is set to True, then the Sobol measure for each
    parameter will be given as its sign the sign of the average gradient with respect to
    that parameter across the search space. Thus, important parameters that, when
    increased, decrease f will have large and negative values; unimportant parameters
    will have values close to 0.

    Args:
        adapter: A Adapter object with models that were fit.
        metrics: The names of the metrics and outcomes for which to compute
            sensitivities. This should preferably be metrics with a good model fit.
            Defaults to adapter.outcomes.
        order: A string specifying the order of the Sobol indices to be computed.
            Supports "first" and "total" and defaults to "first".
        signed: A bool for whether the measure should be signed.
        sobol_kwargs: keyword arguments passed on to SobolSensitivityGPMean, and if
            signed, GpDGSMGpMean.

    Returns:
        Dictionary {'metric_name': {'parameter_name' or
        (parameter_name_1, 'parameter_name_2'): sensitivity_value}}, where the
        `sensitivity` value is cast to a Numpy array in order to be compatible with
        `plot_feature_importance_by_feature`.
    """
    if order not in ["first", "total", "second"]:
        raise NotImplementedError(
            f"Order {order} is not supported. Plese choose one of"
            " 'first', 'total' or 'second'."
        )
    if metrics is None:
        metrics = adapter.outcomes
    generator, digest = _get_generator_and_digest(adapter=adapter)
    model_list = _get_model_per_metric(generator=generator, metrics=metrics)

    # get device and dtype of the first model
    first_model = next(model_list[0].parameters())
    device = first_model.device
    bounds = torch.tensor(
        digest.bounds,
        device=device,
    ).T  # transposing to make it 2 x d

    # for second order indices, we need to compute first order indices first
    # which is what is done here. With the first order indices, we can then subtract
    # appropriately using the first-order indices to extract the second-order indices.
    ind = compute_sobol_indices_from_model_list(
        model_list=model_list,
        bounds=bounds,
        order="first" if order == "second" else order,
        discrete_features=digest.categorical_features + digest.ordinal_features,
        **sobol_kwargs,
    )
    feature_names = digest.feature_names
    if signed:
        ind_deriv = compute_derivatives_from_model_list(
            model_list=model_list,
            bounds=bounds,
            discrete_features=digest.categorical_features + digest.ordinal_features,
            **sobol_kwargs,
        )
        # categorical features don't have a direction, so we set the derivative to 1.0
        # in order not to zero our their sensitivity. We treat categorical features
        # separately in the sensitivity analysis plot as well, to make clear that they
        # are affecting the metric, but neither increasing nor decreasing. Note that the
        # orginal variables have a well defined direction, so we do not need to treat
        # them differently here.
        for i in digest.categorical_features:
            ind_deriv[:, i] = 1.0
        ind *= torch.sign(ind_deriv).to(device)

    indices = array_with_string_indices_to_dict(
        rows=metrics, cols=feature_names, A=ind.cpu().numpy()
    )
    if order == "second":
        second_order_values = compute_sobol_indices_from_model_list(
            model_list=model_list,
            bounds=bounds,
            order="second",
            discrete_features=digest.categorical_features + digest.ordinal_features,
            **sobol_kwargs,
        )
        second_order_feature_names = [
            f"{f1} & {f2}" for f1, f2 in itertools.combinations(digest.feature_names, 2)
        ]

        second_order_dict = array_with_string_indices_to_dict(
            rows=metrics,
            cols=second_order_feature_names,
            A=second_order_values.cpu().numpy(),
        )
        for metric in metrics:
            indices[metric].update(second_order_dict[metric])

    return indices


def _get_generator_and_digest(
    adapter: TorchAdapter,
) -> tuple[LegacyBoTorchGenerator | ModularBoTorchGenerator, SearchSpaceDigest]:
    """Returns the generator of the adapter and the SearchSpaceDigest
    that was used to fit the adapter.
    """
    if not isinstance(adapter, TorchAdapter):
        raise NotImplementedError(
            f"{type(adapter)=}, but only TorchAdapter is supported."
        )
    generator = adapter.generator
    if not isinstance(generator, (LegacyBoTorchGenerator, ModularBoTorchGenerator)):
        raise NotImplementedError(
            f"{type(generator)=}, but only LegacyBoTorchGenerator and "
            "ModularBoTorchGenerator are supported."
        )
    return generator, generator.search_space_digest


def _get_model_per_metric(
    generator: LegacyBoTorchGenerator | ModularBoTorchGenerator, metrics: list[str]
) -> list[GPyTorchModel]:
    """For a given TorchGenerator model, returns a list of botorch.models.model.Model
    objects corresponding to - and in the same order as - the given metrics.
    """
    if isinstance(generator, LegacyBoTorchGenerator):
        # guaranteed not to be None after accessing search_space_digest
        gp_model = generator.model
        model_idx = [generator.metric_names.index(m) for m in metrics]
        if not isinstance(gp_model, ModelList):
            if gp_model.num_outputs == 1:  # can accept single output models
                return [assert_is_instance(gp_model, GPyTorchModel) for _ in model_idx]
            raise NotImplementedError(
                f"type(adapter.generator.model) = {type(gp_model)}, "
                "but only ModelList is supported."
            )
        return [gp_model.models[i] for i in model_idx]
    else:  # isinstance(model, ModularBoTorchGenerator):
        surrogate = generator.surrogate
        outcomes = surrogate.outcomes
        model_list = []
        for m in metrics:  # for each metric, find a corresponding surrogate
            i = outcomes.index(m)
            metric_model = surrogate.model
            # since model is a ModularBoTorchGenerator, metric_model will be a
            # `botorch.models.model.Model` object, which have the `num_outputs`
            # property and `subset_outputs` method.
            if metric_model.num_outputs > 1:  # subset to relevant output
                metric_model = metric_model.subset_output([i])
            model_list.append(metric_model)
        return model_list


def array_with_string_indices_to_dict(
    rows: list[str],
    cols: list[str],
    A: npt.NDArray,
) -> dict[str, dict[str, npt.NDArray]]:
    """
    Args:
        rows: A list of strings with which to index rows of A.
        cols: A list of strings with which to index columns of A.
        A: A matrix, with `len(rows)` rows and `len(cols)` columns.

    Returns:
        A dictionary dict that satisfies dict[rows[i]][cols[j]] = A[i, j].
    """
    return {r: dict(zip(cols, a)) for r, a in zip(rows, A)}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


# Scheduler was deprecated in Ax 1.1.0, so it should be removed in
# Ax 1.2.0+.
class Scheduler:
    raise DeprecationWarning(
        "Scheduler is deprecated following renaming of the Scheduler to "
        "Orchestrator. Please use Orchestrator instead; import with: "
        "`from ax.service.orchestrator import Orchestrator`"
    )


# Scheduler was deprecated in Ax 1.1.0, so it should be removed in
# Ax 1.2.0+.
class SchedulerInternalError(Exception):
    raise DeprecationWarning(
        "SchedulerInternalError is deprecated following renaming of the Scheduler to "
        "Orchestrator. Please use OrchestratorInternalError instead; import with: "
        "`from ax.service.orchestrator import OrchestratorInternalError`"
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCardGroup
from ax.analysis.plotly.cross_validation import CrossValidationPlot
from ax.core.experiment import Experiment
from ax.core.utils import is_bandit_experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import none_throws, override

DIAGNOSTICS_CARDGROUP_TITLE = "Diagnostic Analysis"

DIAGNOSTICS_CARDGROUP_SUBTITLE = (
    "Diagnostic Analyses provide information about the optimization process and "
    "the quality of the model fit. You can use this information to understand "
    "if the experimental design should be adjusted to improve optimization quality."
)


class DiagnosticAnalysis(Analysis):
    """
    An Analysis that provides diagnostic information about the optimization process.
    This includes information about the quality of the model fit, such as the results
    of leave-one-out cross validation.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCardGroup:
        if experiment is None:
            raise UserInputError("DiagnosticAnalysis requires an Experiment.")

        # Extract all metric names from the OptimizationConfig.
        metric_names = [*none_throws(experiment.optimization_config).metrics.keys()]

        is_bandit = generation_strategy and is_bandit_experiment(
            generation_strategy_name=generation_strategy.name
        )

        cross_validation_plots = (
            [
                CrossValidationPlot(metric_names=metric_names).compute_or_error_card(
                    experiment=experiment,
                    generation_strategy=generation_strategy,
                    adapter=adapter,
                )
            ]
            if not is_bandit
            else []
        )

        return self._create_analysis_card_group(
            title=DIAGNOSTICS_CARDGROUP_TITLE,
            subtitle=DIAGNOSTICS_CARDGROUP_SUBTITLE,
            children=[*cross_validation_plots],
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

# Scheduler was deprecated in Ax 1.1.0, so it should be removed in Ax
# Ax 1.2.0+.


class SchedulerOptions:
    raise DeprecationWarning(
        "Scheduler is deprecated following renaming of the Scheduler to "
        "Orchestrator. Please use Orchestrator and OrchestratorOptions instead; "
        "import with: `from ax.service.orchestrator import Orchestrator, "
        "OrchestratorOptions`"
    )
